__all__ = [
    "BaseDataConnection"
]

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import io
import json
import os

from abc import ABC, abstractmethod
from typing import Union, Tuple, TYPE_CHECKING
from warnings import warn

import pandas as pd

import ibm_watson_machine_learning._wrappers.requests as requests
from pandas import concat, DataFrame, read_csv

from ibm_watson_machine_learning.utils.autoai.utils import try_load_dataset, try_load_tar_gz, try_import_autoai_libs
from ibm_watson_machine_learning.utils.autoai.enums import DataConnectionTypes
from ibm_watson_machine_learning.utils.autoai.errors import CannotReadSavedRemoteDataBeforeFit, WrongAssetType
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure, WMLClientError, DataStreamError
from ibm_watson_machine_learning.utils.utils import is_lib_installed
from ibm_watson_machine_learning.data_loaders.datasets.experiment import DEFAULT_SAMPLING_TYPE, DEFAULT_SAMPLE_SIZE_LIMIT

if TYPE_CHECKING:
    from ibm_boto3 import resource


class BaseDataConnection(ABC):
    """Base class for DataConnection."""

    def __init__(self):

        self.type = None
        self.connection = None
        self.location = None
        self.auto_pipeline_params = None
        self._wml_client = None
        self._run_id = None
        self._obm = None
        self._obm_cos_path = None
        self.id = None
        self._datasource_type = None

    @abstractmethod
    def _to_dict(self) -> dict:
        """Convert DataConnection object to dictionary representation."""
        pass

    @classmethod
    @abstractmethod
    def _from_dict(cls, _dict: dict) -> 'BaseDataConnection':
        """Create a DataConnection object from dictionary."""
        pass

    @abstractmethod
    def read(self, with_holdout_split: bool = False) -> Union['DataFrame', Tuple['DataFrame', 'DataFrame']]:
        """Download dataset stored in remote data storage."""
        pass

    @abstractmethod
    def write(self, data: Union[str, 'DataFrame'], remote_name: str) -> None:
        """Upload file to a remote data storage."""
        pass

    def _fill_experiment_parameters(self, prediction_type: str, prediction_column: str, holdout_size: float,
                                    csv_separator: str = ',', excel_sheet: Union[str, int] = None,
                                    encoding: str = 'utf-8') -> None:
        """To be able to recreate a holdout split, this method need to be called."""
        self.auto_pipeline_params = {
            'prediction_type': prediction_type,
            'prediction_column': prediction_column,
            'holdout_size': holdout_size,
            'csv_separator': csv_separator,
            'excel_sheet': excel_sheet,
            'encoding': encoding
        }

    def _download_obm_data_from_cos(self, cos_client: 'resource') -> 'DataFrame':
        """Download preprocessed OBM data. COS version."""

        if '//' in self._obm_cos_path:  # sometimes there is an additional slash, need to replace it
            self._obm_cos_path = self._obm_cos_path.replace('//', '/')

        # note: fetch all OBM file part names
        cos_summary = cos_client.Bucket(self.location.bucket).objects.filter(Prefix=self._obm_cos_path)
        file_names = [file_name.key for file_name in cos_summary]

        # note: if path does not exist, try to find in different one
        if not file_names:
            cos_summary = cos_client.Bucket(self.location.bucket).objects.filter(
                Prefix=self._obm_cos_path.split('./')[-1])
            file_names = [file_name.key for file_name in cos_summary]
            # --- end note
        # --- end note

        # TODO: this can be done simultaneously (multithreading / multiprocessing)
        # note: download all data parts and concatenate them into one output
        parts = []
        for file_name in file_names:
            file = cos_client.Object(self.location.bucket, file_name).get()
            buffer = io.BytesIO(file['Body'].read())
            parts.append(try_load_dataset(buffer=buffer))

        data = concat(parts)
        # --- end note
        return data

    def _download_obm_json(self) -> dict:
        """Download obm.json log."""
        if self._obm:
            if self._obm_cos_path:
                path = self._obm_cos_path.rsplit('features', 1)[0] + 'obm.json'
            else:
                path = f"{self.location.path}/{self._run_id}/data/obm/obm.json"
            data = self._download_json_file(path)
            return data
        else:
            raise Exception("OBM function called when not OBM scenario.")

    def _download_csv_file(self, path) -> dict:
        """Download csv file."""
        df = pd.DataFrame()

        if '//' in path:  # sometimes there is an additional slash, need to replace it
            path = path.replace('//', '/')

        if self.type == DataConnectionTypes.FS:
            csv_url = self._wml_client.service_instance._href_definitions.get_wsd_model_attachment_href() + f"/auto_ml/{path.split('/auto_ml/')[-1]}"

            with requests.get(csv_url, params=self._wml_client._params(), headers=self._wml_client._get_headers(),
                              stream=True) as file_response:
                if file_response.status_code != 200:
                    raise ApiRequestFailure(u'Failure during {}.'.format("downloading json"), file_response)

                downloaded_asset = file_response.content
                df = pd.read_csv(io.BytesIO(downloaded_asset))
                #json_content = json.loads(buffer.getvalue().decode('utf-8'))

        elif self.type == DataConnectionTypes.CA or self.type == DataConnectionTypes.CN:
            if self._check_if_connection_asset_is_s3():
                cos_client = self._init_cos_client()

                try:
                    file = cos_client.Object(self.location.bucket, path).get()
                    content = file["Body"].read()
                    df = pd.read_csv(io.BytesIO(content))

                except Exception as cos_access_exception:
                    raise ConnectionError(
                        f"Unable to access data object in cloud object storage with credentials supplied. "
                        f"Error: {cos_access_exception}")

            else:
                raise NotImplementedError(f"Unsupported connection type: {self.type}. "
                                          f"Datasource type is not supported. "
                                          f"Supported type is: bluemixcloudobjectstorage")

        else:
            raise NotImplementedError(f"Unsupported connection type: {self.type}")

        return df

    def _download_json_file(self, path) -> dict:
        """Download json file."""
        json_content = {}

        if '//' in path:  # sometimes there is an additional slash, need to replace it
            path = path.replace('//', '/')

        # TODO: remove S3 implementation
        if self.type == DataConnectionTypes.S3:
            warn(message="S3 DataConnection is deprecated! Please use data_asset_id instead.")

            cos_client = self._init_cos_client()

            try:
                file = cos_client.Object(self.location.bucket, path).get()
                content = file["Body"].read()
                json_content = json.loads(content.decode('utf-8'))
            except Exception as cos_access_exception:
                raise ConnectionError(
                    f"Unable to access data object in cloud object storage with credentials supplied. "
                    f"Error: {cos_access_exception}")
        elif self.type == DataConnectionTypes.FS:
            json_url = self._wml_client.service_instance._href_definitions.get_wsd_model_attachment_href() + f"/auto_ml/{path.split('/auto_ml/')[-1]}"

            with requests.get(json_url, params=self._wml_client._params(), headers=self._wml_client._get_headers(),
                              stream=True) as file_response:
                if file_response.status_code != 200:
                    raise ApiRequestFailure(u'Failure during {}.'.format("downloading json"), file_response)

                downloaded_asset = file_response.content
                buffer = io.BytesIO(downloaded_asset)
                json_content = json.loads(buffer.getvalue().decode('utf-8'))

        elif self.type == DataConnectionTypes.CA or self.type == DataConnectionTypes.CN:
            if self._check_if_connection_asset_is_s3():
                cos_client = self._init_cos_client()

                try:
                    file = cos_client.Object(self.location.bucket, path).get()
                    content = file["Body"].read()
                    json_content = json.loads(content.decode('utf-8'))

                except Exception as cos_access_exception:
                    raise ConnectionError(
                        f"Unable to access data object in cloud object storage with credentials supplied. "
                        f"Error: {cos_access_exception}")

            else:
                raise NotImplementedError(f"Unsupported connection type: {self.type}. "
                                          f"Datasource type is not supported. "
                                          f"Supported type is: bluemixcloudobjectstorage")

        else:
            raise NotImplementedError(f"Unsupported connection type: {self.type}")

        return json_content

    def _get_attachemnt_details(self, data_asset_id: str):
        response = requests.get(self._wml_client.service_instance._href_definitions.get_data_asset_href(data_asset_id),
                                params=self._wml_client._params(),
                                headers=self._wml_client._get_headers())
        data_asset_details = self._wml_client.data_assets._handle_response(200, u'Cannot get data asset details', response)

        attachments_data_asset_details = data_asset_details.get('attachments', [{}])[0]
        # note: Return attachment details from data asset details if it is not a connected data asset:
        if attachments_data_asset_details and attachments_data_asset_details.get('connection_id') is None:
            return attachments_data_asset_details
        else:
            attachment_id = attachments_data_asset_details.get('id')
            if not self._wml_client.ICP and not self._wml_client.WSD:
                response = requests.get(
                    self._wml_client.service_instance._href_definitions.get_attachment_href(data_asset_id, attachment_id),
                    params=self._wml_client._params(),
                    headers=self._wml_client._get_headers())
            else:
                response = requests.get(
                    self._wml_client.service_instance._href_definitions.get_attachment_href(data_asset_id, attachment_id),
                    params=self._wml_client._params(),
                    headers=self._wml_client._get_headers(),
                    verify=False)

            self._wml_client.data_assets._handle_response(200, u'Cannot get attachment details', response)

            return response.json()

    def _check_if_connection_asset_is_s3(self) -> bool:
        try:
            # fast return true if the connection is an instance of S3Connection and non s3 type, or its attribute
            # "cos_type" takes True
            from .connections import S3Connection
            if (isinstance(self.connection, S3Connection) and self.type != 's3' and self.connection.to_dict()) or (
                    self.connection is not None and getattr(self.connection, 'cos_type', False)):
                return True

            if self.type == 'data_asset':
                if self.connection is not None and hasattr(self.connection, 'href'):
                    items = self.connection.href.split('/')

                else:
                    items = self.location.href.split('/')
                _id = items[-1].split('?')[0]

                if self._wml_client is not None:
                    attachment_details = self._get_attachemnt_details(_id)
                    if 'connection_id' not in attachment_details:
                        if 'handle' in attachment_details and 'bucket' in attachment_details['handle']:
                            connection_details = self._create_conn_details_for_container()
                            attachment_details['connection_path'] = '/' + attachment_details['handle']['bucket'] + '/' + attachment_details['handle']['key']
                        elif 'url' in attachment_details and 'name' in attachment_details and 'bucket' in attachment_details:
                            connection_details = self._create_conn_details_for_container()
                            attachment_details['connection_path'] = '/' + attachment_details['bucket']['bucket_name'] + '/' + attachment_details['name']
                        else:
                            return False

                    else:
                        connection_details = self._wml_client.connections.get_details(
                            attachment_details['connection_id'])

                else:
                    try:
                        from project_lib import Project
                        project = Project.access()

                        # note: Check if asset is located directly in the project files
                        #       If yes it is not a connected data.
                        #       [Prevents unnecessary logging].
                        if any(file['asset_id'] == _id for file in project.get_files()):
                            return False
                        # --- end note

                        try:
                            connection_details = project.get_connected_data(_id)

                        except RuntimeError:  # when data asset is normal not s3 or database
                            return False

                        attachment_details = {'connection_path': connection_details.get('datapath')}

                    except ModuleNotFoundError:
                        raise NotImplementedError(f"This functionality can be run only on Watson Studio.")

            elif self.type == 'connection_asset':
                if self._wml_client is not None:
                    connection_details = self._wml_client.connections.get_details(self.connection.id)

                else:
                    try:
                        from project_lib import Project
                        project = Project.access()
                        connection_details = project.get_connection(self.connection.id)

                    except ModuleNotFoundError:
                        raise NotImplementedError(f"This functionality can be run only on Watson Studio.")

            elif self.type == 'container':
                connection_details = self._create_conn_details_for_container()

            elif self.type == 's3' or self.type == 'fs':
                return False

            else:
                raise WrongAssetType(asset_type=self.type)

            # Note: Check with project libs if connection points to S3
            if self._wml_client is not None:
                datasource_type = connection_details['entity']['datasource_type']
                self._datasource_type = datasource_type
                datasource_type_id_ibm_cos = self._wml_client.connections.get_datasource_type_uid_by_name(
                    'bluemixcloudobjectstorage')
                datasource_type_id_aws_cos = self._wml_client.connections.get_datasource_type_uid_by_name(
                    'cloudobjectstorage')

                if (datasource_type == datasource_type_id_ibm_cos or datasource_type == datasource_type_id_aws_cos or
                        datasource_type == 'bluemixcloudobjectstorage' or datasource_type == 'cloudobjectstorage'):
                    cos_type = True

                else:
                    cos_type = False

            elif self.type == 'container':
                cos_type = True

            elif 'url' in connection_details:
                cos_type = True
                connection_details['entity'] = {'properties': connection_details}

            else:
                cos_type = False
            # --- end note

            if cos_type:
                if 'cos_hmac_keys' in str(connection_details['entity']['properties']):
                    creds = json.loads(connection_details['entity']['properties']['credentials'])
                    connection_details['entity']['properties']['access_key'] = creds['cos_hmac_keys']['access_key_id']
                    connection_details['entity']['properties']['secret_key'] = creds['cos_hmac_keys'][
                        'secret_access_key']

                if self.connection is None:
                    from .connections import S3Connection
                    if ('api_key' in str(connection_details['entity']['properties']) and
                            'iam_url' in str(connection_details['entity']['properties'])):
                        self.connection = S3Connection(
                            api_key=connection_details['entity']['properties'].get('api_key'),
                            auth_endpoint=connection_details['entity']['properties'].get('iam_url'),
                            endpoint_url=connection_details['entity']['properties'].get('url'),
                            resource_instance_id=connection_details['entity']['properties'].get('resource_instance_id'),
                            _internal_use=True
                        )

                    else:
                        self.connection = S3Connection(
                            endpoint_url=connection_details['entity']['properties'].get('url'),
                            access_key_id=connection_details['entity']['properties'].get('access_key'),
                            secret_access_key=connection_details['entity']['properties'].get('secret_key'),
                            _internal_use=True
                        )

                else:
                    if ('api_key' in str(connection_details['entity']['properties']) and
                            'iam_url' in str(connection_details['entity']['properties'])):
                        self.connection.api_key = connection_details['entity']['properties'].get('api_key')
                        self.connection.auth_endpoint = connection_details['entity']['properties'].get('iam_url')
                        self.connection.endpoint_url = connection_details['entity']['properties'].get('url')
                        self.connection.resource_instance_id = connection_details['entity']['properties'].get('resource_instance_id')

                    else:
                        self.connection.secret_access_key = connection_details['entity']['properties'].get('secret_key')
                        self.connection.access_key_id = connection_details['entity']['properties'].get('access_key')
                        self.connection.endpoint_url = connection_details['entity']['properties'].get('url')

                if self.type == 'container':
                    self.location.bucket = connection_details['entity']['properties']['bucket_name']

                if self.type == 'data_asset':
                    try:
                        # Workaround for connected data assets promoted from catalog
                        if not attachment_details['connection_path'].startswith('/'):
                            attachment_details['connection_path'] = '/' + attachment_details['connection_path']
                        self.location.bucket = attachment_details['connection_path'].split('/')[1]

                        # Remove excel sheet name from connection_path if data asset is excel file:
                        if '.xlsx/' in attachment_details['connection_path'] or '.xls/' in attachment_details['connection_path']:
                            excel_sheet = attachment_details['connection_path'].split('/')[-1]
                            # note: set recognized excel sheet in params
                            if not self.auto_pipeline_params.get('excel_sheet'):
                                self.auto_pipeline_params['excel_sheet'] = excel_sheet
                            # end note
                            self.location.path = '/'.join(attachment_details['connection_path'].split('/')[2:-1])
                        else:
                            self.location.path = '/'.join(attachment_details['connection_path'].split('/')[2:])

                    except IndexError:
                        self.location.bucket = connection_details['entity']['properties']['bucket']
                        self.location.path = attachment_details['connection_path']

                    self.location.id = None
                self.connection.cos_type = True
                return True

            else:
                return False

        except Exception as e:
            if os.environ.get('USER_ACCESS_TOKEN') is None and os.environ.get('RUNTIME_ENV_ACCESS_TOKEN_FILE') is None:
                raise e

            else:
                return False  # if we are in WS, ignore this check even if there was some error

    def _is_data_asset_normal(self) -> bool:
        """Returns `True` if data asset is normal data asset - not connected data asset."""
        try:
            if self.type == 'data_asset':
                if self.connection is not None and hasattr(self.connection, 'href'):
                    items = self.connection.href.split('/')
                else:
                    items = self.location.href.split('/')
                _id = items[-1].split('?')[0]

                if self._wml_client is not None:
                    attachment_details = self._get_attachemnt_details(_id)
                    return bool('connection_id' not in attachment_details)

                else:
                    try:
                        from project_lib import Project
                        project = Project.access()

                        # note: Check if asset is located directly in the project files
                        #       If yes it is not a connected data.
                        #       [Prevents unnecessary logging].
                        return any(file['asset_id'] == _id for file in project.get_files())
                        # --- end note

                    except ModuleNotFoundError:
                        raise NotImplementedError(f"This functionality can be run only on Watson Studio.")

        except Exception as e:
            if os.environ.get('USER_ACCESS_TOKEN') is None and os.environ.get('RUNTIME_ENV_ACCESS_TOKEN_FILE') is None:
                raise e
            else:
                return False  # if we are in WS, ignore this check even if there was some error

    def _is_data_asset_nfs(self):
        if self.type == 'data_asset':
            if self.connection is not None and hasattr(self.connection, 'href'):
                items = self.connection.href.split('/')
            else:
                items = self.location.href.split('/')
            _id = items[-1].split('?')[0]

            if self._wml_client is not None:
                attachment_details = self._get_attachemnt_details(_id)

                return ('connection_id' in attachment_details and attachment_details.get(
                    'datasource_type') == self._wml_client.connections.get_datasource_type_uid_by_name('volumes'))
        return False


    def _download_indices_from_cos(self, cos_client: 'resource', location_path) -> 'DataFrame':
        """Download indices for this connection. COS version"""

        try:
            file = cos_client.Object(self.location.bucket,
                                     location_path).get()
        except:
            file = list(cos_client.Bucket(self.location.bucket).objects.filter(
                Prefix=location_path))[0].get()

        buffer = io.BytesIO(file['Body'].read())

        if '.csv' in location_path:
            file_name = 'indices.csv'
            with open(file_name, 'wb') as out:
                out.write(buffer.read())

            data = read_csv(file_name, sep=self.auto_pipeline_params.get('csv_separator', ','),
                            encoding=self.auto_pipeline_params.get('encoding', 'utf-8'))

        else:
            data = try_load_tar_gz(buffer=buffer,
                                   separator=self.auto_pipeline_params.get('csv_separator', ','),
                                   encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                   )

        return data

    def _download_training_data_from_data_asset_storage(self) -> 'DataFrame':
        """Download training data for this connection. Data Storage."""

        if self._wml_client is not None:
            # note: as we need to load a data into the memory,
            # we are using pure requests and helpers from the WML client
            asset_id = self.location.href.split('?')[0].split('/')[-1]

            # note: download data asset details
            asset_response = requests.get(
                self._wml_client.service_instance._href_definitions.get_data_asset_href(asset_id),
                params=self._wml_client._params(),
                headers=self._wml_client._get_headers())

            asset_details = self._wml_client.data_assets._handle_response(200, u'get assets', asset_response)

            attachment_id = asset_details['attachments'][0]['id']
            response = requests.get(
                self._wml_client.service_instance._href_definitions.get_attachment_href(asset_id, attachment_id),
                params=self._wml_client.data_assets._client._params(),
                headers=self._wml_client.data_assets._client._get_headers())

            if response.status_code == 200:
                try:
                    attachment_details = response.json()
                    if "url" in attachment_details:
                        file_asset_url = attachment_details['url']
                        if not file_asset_url.startswith("http"):
                            file_asset_url = self._wml_client.wml_credentials['url'] + file_asset_url
                        csv_response = requests.get(file_asset_url)

                        if csv_response.status_code != 200:
                            raise ApiRequestFailure(u'Failure during {}.'.format("downloading model"), csv_response)

                        downloaded_asset = csv_response.content

                        # note: read the csv/xlsx file from the memory directly into the pandas DataFrame
                        buffer = io.BytesIO(downloaded_asset)
                        data = try_load_dataset(buffer=buffer,
                                                sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                                separator=self.auto_pipeline_params.get('csv_separator', ','),
                                                encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                                )

                        return data

                except requests.exceptions.MissingSchema:
                    pass  # go to 'handle' part and check if we are able to download data asset from WS

            # note: read the csv url
            if 'handle' in asset_details['attachments'][0]:
                attachment_url = asset_details['attachments'][0]['handle']['key']

                # note: make the whole url pointing out the csv
                artifact_content_url = (
                    f"{self._wml_client.service_instance._href_definitions.get_wsd_model_attachment_href()}"
                    f"{attachment_url}")

                # note: stream the whole CSV file
                csv_response = requests.get(artifact_content_url,
                                            params=self._wml_client._params(),
                                            headers=self._wml_client._get_headers(),
                                            stream=True)

                if csv_response.status_code != 200:
                    raise ApiRequestFailure(u'Failure during {}.'.format("downloading model"), csv_response)

                downloaded_asset = csv_response.content

                # note: read the csv/xlsx file from the memory directly into the pandas DataFrame
                buffer = io.BytesIO(downloaded_asset)
                data = try_load_dataset(buffer=buffer,
                                        sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                        separator=self.auto_pipeline_params.get('csv_separator', ','),
                                        encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                        )

                return data

            else:
                # NFS scenario
                connection_id = asset_details['attachments'][0]['connection_id']
                connection_path = asset_details['attachments'][0]['connection_path']

                return self._download_data_from_nfs_connection_using_id_and_path(connection_id, connection_path)

        else:
            try:
                from project_lib import Project
                project = Project.access()

            except ModuleNotFoundError:
                raise NotImplementedError(f"This functionality can be run only on Watson Studio.")

            asset_id = self.location.href.split('?')[0].split('/')[-1]
            assets_list = project.get_assets()

            data_asset_name = None
            for asset in assets_list:
                if asset['asset_id'] == asset_id:
                    data_asset_name = asset['name']

            if data_asset_name is None:
                raise FileNotFoundError(f"Cannot find data asset with id: {asset_id}")

            buffer = project.get_file(data_asset_name)
            data = try_load_dataset(buffer=buffer,
                                    sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                    separator=self.auto_pipeline_params.get('csv_separator', ','),
                                    encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                    )

            return data

    def _download_training_data_from_file_system(self, binary: bool = False) -> Tuple['DataFrame', bytes]:
        """Download training data for this connection. File system version."""

        try:
            url = self._wml_client.service_instance._href_definitions.get_wsd_model_attachment_href() + f"/{self.location.path.split('/assets/')[-1]}"
            # note: stream the whole CSV file
            csv_response = requests.get(url,
                                        params=self._wml_client._params(),
                                        headers=self._wml_client._get_headers(),
                                        stream=True)

            if csv_response.status_code != 200:
                raise ApiRequestFailure(u'Failure during {}.'.format("downloading model"), csv_response)

            downloaded_asset = csv_response.content

            if binary:
                return downloaded_asset

            # note: read the csv/xlsx file from the memory directly into the pandas DataFrame
            buffer = io.BytesIO(downloaded_asset)
            data = try_load_dataset(buffer=buffer,
                                    sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                    separator=self.auto_pipeline_params.get('csv_separator', ','),
                                    encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                    )
        except (ApiRequestFailure, AttributeError):
            with open(self.location.path, 'rb') as data_buffer:
                if binary:
                    return data_buffer.read()

                data = try_load_dataset(buffer=data_buffer,
                                        sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                        separator=self.auto_pipeline_params.get('csv_separator', ','),
                                        encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                        )

        return data

    def _download_indices_from_file_system(self, location_path: str) -> 'DataFrame':
        """Download indices for this connection. File system version."""

        try:
            url = self._wml_client.service_instance._href_definitions.get_wsd_model_attachment_href() + f"/{location_path.split('/assets/')[-1]}"
            # note: stream the whole CSV file
            csv_response = requests.get(url,
                                        params=self._wml_client._params(),
                                        headers=self._wml_client._get_headers(),
                                        stream=True)

            if csv_response.status_code != 200:
                raise ApiRequestFailure(u'Failure during {}.'.format("downloading model"), csv_response)

            downloaded_asset = csv_response.content
            # note: read the csv/xlsx file from the memory directly into the pandas DataFrame
            buffer = io.BytesIO(downloaded_asset)
            data = try_load_dataset(buffer=buffer,
                                    sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                    separator=self.auto_pipeline_params.get('csv_separator', ','),
                                    encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                    )
        except (ApiRequestFailure, AttributeError):
            with open(location_path, 'rb') as data_buffer:
                data = try_load_tar_gz(buffer=data_buffer,
                                       separator=self.auto_pipeline_params.get('csv_separator', ','),
                                       encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                       )

        return data

    def _download_obm_data_from_file_system(self) -> 'DataFrame':
        """Download preprocessed OBM data. FS version."""

        # note: fetch all OBM file part names
        url = self._wml_client.service_instance._href_definitions.get_wsd_model_attachment_href() + f"/auto_ml/{self.location.path.split('/auto_ml/')[-1]}/{self._run_id}/data/obm/features"
        params = self._wml_client._params()
        params['flat'] = "true"

        response = requests.get(url,
                                params=params,
                                headers=self._wml_client._get_headers())

        if response.status_code != 200:
            raise ApiRequestFailure(u'Failure during {}.'.format("getting files information"), response)

        file_names = [e['path'].split('/')[-1] for e in response.json()['resources'] if
                      e['type'] == 'file' and e['path'].split('/')[-1].startswith('part')]

        # TODO: this can be done simultaneously (multithreading / multiprocessing)
        # note: download all data parts and concatenate them into one output
        parts = []
        for file_name in file_names:
            csv_response = requests.get(url + '/' + file_name,
                                        params=self._wml_client._params(),
                                        headers=self._wml_client._get_headers(),
                                        stream=True)

            if csv_response.status_code != 200:
                raise ApiRequestFailure(u'Failure during {}.'.format("downloading model"), csv_response)

            downloaded_asset = csv_response.content
            # note: read the csv/xlsx file from the memory directly into the pandas DataFrame
            buffer = io.BytesIO(downloaded_asset)
            parts.append(try_load_dataset(buffer=buffer))

        data = concat(parts)
        # --- end note
        return data

    def _download_data_from_nfs_connection(self) -> 'DataFrame':
        """Download training data for this connection. NFS."""

        # note: as we need to load a data into the memory,
        # we are using pure requests and helpers from the WML client
        data_path = self.location.path
        connection_id = self.connection.asset_id

        return self._download_data_from_nfs_connection_using_id_and_path(connection_id, data_path)

    def _download_data_from_nfs_connection_using_id_and_path(self, connection_id, connection_path) -> 'DataFrame':
        """Download training data for this connection. NFS."""

        # it means that it is on ICP env and it is before fit, so let's throw error
        if not self._wml_client:
            raise CannotReadSavedRemoteDataBeforeFit()

        buffer = None
        # Note: workaround with volumes API as connections API changes data format
        try:
            connection_details = self._wml_client.connections.get_details(connection_id)
        except ApiRequestFailure as conn_error:
            if os.environ.get('TRAINING_NFS_PATH'):
                # Note: Only viable on AutoAI runtime
                base_path = os.environ.get('TRAINING_NFS_PATH') + "/0"
                if connection_path.startswith('/'):
                    data_path = f"{base_path}{connection_path}"
                else:
                    data_path = f"{base_path}/{connection_path}"
                with open(data_path, 'rb') as f:
                    buffer = io.BytesIO(f.read())
            else:
                raise conn_error

        if buffer is None:
            href = self._wml_client.volumes._client.service_instance._href_definitions.volume_upload_href(
                connection_details['entity']['properties']['volume'])
            full_href = f"{href[:-1]}{connection_path}"

            csv_response = requests.get(full_href,
                                        headers=self._wml_client._get_headers(),
                                        stream=True)

            # Note: if file is written in directory we need to create different href for download
            if csv_response.status_code != 200:
                path_parts = connection_path.split('/')
                full_href = f"{href[:-1]}{'/'.join(path_parts[:-1])}%2F{path_parts[-1]}"

                csv_response = requests.get(full_href,
                                            headers=self._wml_client._get_headers(),
                                            stream=True)

                if csv_response.status_code != 200:
                    raise ApiRequestFailure(u'Failure during {}.'.format("downloading data"), csv_response)

            downloaded_asset = csv_response.content
            # note: read the csv/xlsx file from the memory directly into the pandas DataFrame
            buffer = io.BytesIO(downloaded_asset)

        data = try_load_dataset(buffer=buffer,
                                sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                separator=self.auto_pipeline_params.get('csv_separator', ','),
                                encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                )

        return data

    def _download_data_from_cos(self, cos_client: 'resource') -> 'DataFrame':
        """Download training data for this connection. COS version"""
        location = self.location.file_name if hasattr(self.location, "file_name") else self.location.path
        try:
            file = cos_client.Object(self.location.bucket,
                                     location).get()
        except Exception as e:
            cos_objects_list = list(cos_client.Bucket(self.location.bucket).objects.filter(
                Prefix=location))
            if len(cos_objects_list) > 0:
                file = list(cos_client.Bucket(self.location.bucket).objects.filter(
                    Prefix=location))[0].get()
            else:
                raise e

        buffer = io.BytesIO(file['Body'].read())
        data = try_load_dataset(buffer=buffer,
                                sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                separator=self.auto_pipeline_params.get('csv_separator', ','),
                                encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                )

        return data

    def _create_conn_details_for_container(self) -> dict:
        if self._wml_client is not None:
            if self._wml_client.default_space_id is not None:
                details = self._wml_client.spaces.get_details(self._wml_client.default_space_id)

            else:
                project_url = self._wml_client.service_instance._href_definitions.get_project_href(
                    self._wml_client.default_project_id)

                response = requests.get(
                    project_url,
                    headers=self._wml_client._get_headers(projects_token=True)
                )

                details = response.json()

            properties = details['entity']['storage']['properties']
            creds = details['entity']['storage']['properties']['credentials'].get('admin')
            creds = creds if creds else details['entity']['storage']['properties']['credentials'].get('editor')
            properties.update(creds)
            properties['url'] = properties['endpoint_url']
            properties['access_key'] = properties.get('access_key_id')
            properties['secret_key'] = properties.get('secret_access_key')

            connection_details = {
                'entity': {
                    'datasource_type': 'bluemixcloudobjectstorage',
                    'properties': properties
                }
            }

        else:
            try:
                from project_lib import Project
                token = self._get_token_from_environment()

                if token is None:
                    raise NotImplementedError(
                        f"""To succesfully read the training data used in AutoAI experiment, you need to provide the project token.
                    **To insert the project token to your notebook:**
                        Click the More icon on your notebook toolbar and then click Insert project token.
                        Run the inserted code cell.
                    Note:
                    If you are told in a message that no project token exists, click the link in the message to be redirected to the project's Settings page where you can create a project token.
                    **To create a project token:**
                        Click New token in the Access tokens section on the Settings page of your project.
                        Enter a name, select Editor role for the project, and create a token.
                        Go back to your notebook, click the More icon on the notebook toolbar and then click Insert project token.
                        Run the inserted code cell."""
                    )

                token = token.split('Bearer ')[-1]

                _project = Project(project_id=os.environ.get('PROJECT_ID'),
                                   project_access_token=token)

                details = _project.get_storage_metadata()

                properties = details['properties']
                properties.update(properties['credentials']['editor'])
                properties['url'] = properties['endpoint_url']
                properties['access_key'] = properties.get('access_key_id')
                properties['secret_key'] = properties.get('secret_access_key')

                connection_details = {
                    'entity': {
                        'datasource_type': 'bluemixcloudobjectstorage',
                        'properties': properties
                    }
                }

            except ModuleNotFoundError:
                raise NotImplementedError(f"This functionality can be run only on Watson Studio.")

        return connection_details

    def _download_data_from_flight_service(self,
                                           data_location: 'DataConnection',
                                           binary: bool = False,
                                           read_to_file: str = None,
                                           flight_parameters: dict = None,
                                           headers: dict = None,
                                           number_of_batch_rows: int = None,
                                           sampling_type: str = DEFAULT_SAMPLING_TYPE,
                                           return_data_as_iterator: bool = False,
                                           enable_sampling: bool = True,
                                           _return_subsampling_stats: bool = False,
                                           total_size_limit=DEFAULT_SAMPLE_SIZE_LIMIT,
                                           total_nrows_limit=None,
                                           total_percentage_limit=1.0):

        is_lib_installed(lib_name='pyarrow', minimum_version='3.0.0', install=True)

        # import the class only if flight scenario is enabled - do not import it in main import section
        from ibm_watson_machine_learning.data_loaders.experiment import ExperimentDataLoader
        from ibm_watson_machine_learning.data_loaders.datasets.experiment import ExperimentIterableDataset

        if flight_parameters is None:
            flight_parameters = {"num_partitions": 4}

        dict_connection = data_location._to_dict()

        experiment_metadata = {
            "n_parallel_data_connections": flight_parameters.get("num_partitions", 4),
            "prediction_column": self.auto_pipeline_params.get('prediction_column'),
            "prediction_type": self.auto_pipeline_params.get('prediction_type'),
            "project_id": self._wml_client.default_project_id,
            "space_id": self._wml_client.default_space_id,
            'headers': self._wml_client._get_headers() if headers is None else headers
        }

        experiment_metadata.update(self.auto_pipeline_params)

        experiment_iterable_dataset_setup_parameters = dict(
            connection=dict_connection,
            enable_sampling=enable_sampling if not binary else False,  # don't use sampling to read binary
            experiment_metadata=experiment_metadata,
            binary_data=binary,
            read_to_file=read_to_file,
            flight_parameters=flight_parameters if flight_parameters is not None else {},
            fallback_to_one_connection=False,
            _return_subsampling_stats=_return_subsampling_stats,
            number_of_batch_rows=number_of_batch_rows,
            sampling_type=sampling_type,
            _wml_client=self._wml_client,
            total_size_limit=total_size_limit,
            total_nrows_limit=total_nrows_limit,
            total_percentage_limit=total_percentage_limit
        )

        iterable_dataset = ExperimentIterableDataset(**experiment_iterable_dataset_setup_parameters)
        data_loader = ExperimentDataLoader(dataset=iterable_dataset)

        try:

            if not return_data_as_iterator:
                for data in data_loader:
                    return data
            else:
                return data_loader
        except TypeError as e1:
            # note: retry if there is problem with types:
            # try download data with  infer_as_varchar set to 'true' if some error occurs
            # final data downloaded and converted to proper types might has smaller size than sample_size_limit,
            # because data limit is calculated on data downloaded as varchar, before the conversion to more optimal type.
            try:
                iterable_dataset = ExperimentIterableDataset(**experiment_iterable_dataset_setup_parameters)

                iterable_dataset.connection.infer_as_varchar = 'true'
                data_loader = ExperimentDataLoader(dataset=iterable_dataset)

                if not return_data_as_iterator:
                    for data in data_loader:
                        return data
                else:
                    return data_loader
            except Exception as e2:
                raise DataStreamError(f"First attempt of downloading data failed with error: {e1}. \n"
                                      f"Retry and use infer as varchar also failed with error: {e2}.")

    def _upload_data_via_flight_service(self,
                                        data_location: 'DataConnection',
                                        data: DataFrame = None,
                                        file_path: str = None,
                                        remote_name: str = None,
                                        flight_parameters: dict = None,
                                        headers: dict = None):
        is_lib_installed(lib_name='pyarrow', minimum_version='3.0.0', install=True)

        # import the class only if flight scenario is enabled - do not import it in main import section
        from ibm_watson_machine_learning.data_loaders.datasets.experiment import ExperimentIterableDataset

        if flight_parameters is None:
            flight_parameters = {"num_partitions": 1}

        dict_connection = data_location._to_dict()

        if remote_name:
            dict_connection['location']['path'] = self._get_path_with_remote_name(dict_connection, remote_name)
        elif dict_connection.get('location', {}).get('file_name'):
            dict_connection['location']['path'] = dict_connection['location']['file_name']

        experiment_metadata = {
            "n_parallel_data_connections": flight_parameters.get("num_partitions", 1),
            "project_id": self._wml_client.default_project_id,
            "space_id": self._wml_client.default_space_id,
            'headers': self._wml_client._get_headers() if headers is None else headers
        }

        experiment_metadata.update(self.auto_pipeline_params)

        iterable_dataset = ExperimentIterableDataset(
            connection=dict_connection,
            experiment_metadata=experiment_metadata,
            binary_data=True if file_path is not None else False,
            flight_parameters=flight_parameters if flight_parameters is not None else {},
            _wml_client=self._wml_client
        )

        if data is not None:
            try:
                iterable_dataset.write(data=data)

            except Exception as e:
                if 'gRPC message exceeds maximum size' in str(e):
                    raise ValueError(f"Exceeds maximum data size. Please provide data file path "
                                     f"instead of the pandas DataFrame to upload data in binary mode. Error: {e}")

                else:
                    raise e

        else:
            iterable_dataset.write(file_path=file_path)

    @staticmethod
    def _get_path_with_remote_name(dict_connection: dict, remote_name: str) -> str:
        if dict_connection.get('location', {}).get('path'):
            updated_path = dict_connection['location']['path'] + '/' + remote_name
            updated_path = updated_path.replace('//', '/')
        elif dict_connection.get('location', {}).get('file_name'):
            actual_path = dict_connection['location']['file_name']
            last_slash_index = actual_path.rfind('/') if '/' in actual_path else 0

            if '.' in actual_path[last_slash_index:]:
                actual_path = actual_path[:-last_slash_index]

            if actual_path:
                updated_path = actual_path + '/' + remote_name
                updated_path = updated_path.replace('//', '/')
            else:
                updated_path = remote_name
        else:
            updated_path = remote_name

        return updated_path

    @staticmethod
    def _get_token_from_environment():
        if os.environ.get('RUNTIME_ENV_ACCESS_TOKEN_FILE'):
            with open(os.environ.get('RUNTIME_ENV_ACCESS_TOKEN_FILE'), 'r') as f:
                token = f.read()
        else:
            token = os.environ.get('USER_ACCESS_TOKEN')

        if token:
            token = token.replace('Bearer ', '')

        return token

    def _is_size_acceptable(self):
        """
        Checks if data asset size is acceptable to download based on available memory in pod (MEM env variable)/ T-shirt size.
        Returns: True when data asset size is equal or lower than T-shirt limitation or when limitation is not set (outside autoai pod).
                False when data asset size is known and is above supported limit.
                None when data asset size is unknown or data_connection is not data asset type.
        """
        from ibm_watson_machine_learning.utils.autoai.connection import get_max_sample_size_limit

        if self._wml_client and self.type == 'data_asset':
            if self.connection is not None and hasattr(self.connection, 'href'):
                items = self.connection.href.split('/')
            else:
                items = self.location.href.split('/')
            data_asset_id = items[-1].split('?')[0]
            asset_size = self._wml_client.data_assets.get_details(data_asset_id).get('metadata', {}).get("size", 0)

            if asset_size == 0:                                 # data size unknown
                return None
            elif asset_size <= get_max_sample_size_limit():     # data size is within acceptable range
                return True
            elif not os.environ.get('MEM', False):
                return True                                     # no limitation were set
            else:
                return False                                    # data size is above supported limit
        else:
            return None                                         # not a data asset






#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from .swagger_client.api_client import ApiClient
from .swagger_client.rest import ApiException
import os
import ssl

import certifi

try:
    import urllib3
except ImportError:
    raise ImportError('urllib3 is missing')

try:
    # for python3
    from urllib.parse import urlencode
except ImportError:
    # for python2
    from urllib import urlencode
http = urllib3.PoolManager(
    cert_reqs='CERT_REQUIRED',
    ca_certs=certifi.where()
)


class MLApiClient(ApiClient):
    """

    Class extending ApiClient.

    """
    def __init__(self, repository_path):
        super(MLApiClient, self).__init__(repository_path)
        self.repository_path = repository_path

    def download_file(self, path, presigned_url, query_params, header_params):

        if 'DEPLOYMENT_PLATFORM' in os.environ and os.environ['DEPLOYMENT_PLATFORM'] == 'private':
            http = urllib3.PoolManager(
                cert_reqs=ssl.CERT_NONE,
                ca_certs=certifi.where()
            )
        else:
            http = urllib3.PoolManager(
                cert_reqs='CERT_REQUIRED',
                ca_certs=certifi.where()
            )

        tmp_headers = self.default_headers.copy()
        tmp_headers.update(header_params)
        if presigned_url == 'true':
            path = path.replace('%2F', '/')
            tmp_headers = header_params.copy()

        r = http.request(
            'GET',
            '{}'.format(self.host+path, urlencode(query_params)),
            headers=tmp_headers,
            preload_content=False
            )
        if r.status == 200:
            return r
        else:
            raise ApiException(r.status, 'No content for: {}'.format(path))

    def download_file_v4(self, path, presigned_url, query_params, header_params):
        if 'DEPLOYMENT_PLATFORM' in os.environ and os.environ['DEPLOYMENT_PLATFORM'] == 'private':
            http = urllib3.PoolManager(
                cert_reqs=ssl.CERT_NONE,
                ca_certs=certifi.where()
            )
        else:
            http = urllib3.PoolManager(
                cert_reqs='CERT_REQUIRED',
                ca_certs=certifi.where()
            )
        tmp_headers = self.default_headers.copy()
        tmp_headers.update(header_params)
        if presigned_url == 'true':
            path = path.replace('%2F', '/')
        tmp_headers = header_params.copy()
        tmp_headers.update({'Content-Type': 'application/json'})
        tmp_headers.update({'ML-Instance-ID': 'invalid', 'x-wml-internal-switch-to-new-v4': 'true'})

        r = http.request(
            'GET',
            '{}'.format(self.host + path),
            headers=tmp_headers,
            preload_content=False,
            fields=query_params
        )
        if r.status == 200:
            return r
        else:
            raise ApiException(r.status, 'No content for: {}'.format(path))



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

class UnsupportedNodeTypeError(Exception):
    """ Unsupported node type failure. """
    def __init__(self, *args, **kwargs): # real signature unknown
        pass

    @staticmethod # known case of __new__
    def __new__(S, *more): # real signature unknown; restored from __doc__
        """ T.__new__(S, ...) -> a new object with type S, a subtype of T """
        pass


#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import absolute_import

from .artifact import Artifact
from .artifact_reader import ArtifactReader
from ibm_watson_machine_learning.libs.repo.mlrepository.meta_names import MetaNames
from ibm_watson_machine_learning.libs.repo.mlrepository.meta_props import MetaProps
from ibm_watson_machine_learning.libs.repo.mlrepository.model_artifact import ModelArtifact
from .pipeline_artifact import PipelineArtifact
from ibm_watson_machine_learning.libs.repo.mlrepository.scikit_model_artifact import ScikitModelArtifact
from ibm_watson_machine_learning.libs.repo.mlrepository.xgboost_model_artifact import XGBoostModelArtifact
from .wml_experiment_artifact import WmlExperimentArtifact
from .wml_libraries_artifact import WmlLibrariesArtifact
from .wml_runtimes_artifact import WmlRuntimesArtifact
from .hybrid_model_artifact import  HybridModelArtifact

__all__ = ['Artifact', 'ArtifactReader', 'MetaNames', 'MetaProps', 'WmlExperimentArtifact',
           'ModelArtifact', 'PipelineArtifact', 'ScikitModelArtifact', 'XGBoostModelArtifact',
           'WmlLibrariesArtifact', 'WmlRuntimesArtifact', 'HybridModelArtifact']



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import os
import requests
from requests import packages, exceptions
from functools import wraps

verify = os.environ.get('WML_CLIENT_VERIFY_REQUESTS')

def set_verify_for_requests(func):
    @wraps(func)
    def wrapper(*args, **kw):
        global verify

        if verify is not None:
            if verify == 'True':
                kw.update({'verify': True})

            elif verify == 'False':
                kw.update({'verify': False})

            else:
                kw.update({'verify': verify})

        else:
            kw.update({'verify': True})

        try:
            res = func(*args, **kw)

        except OSError as e:

            # User can pass verify the path to a CA_BUNDLE file or directory with certificates of trusted CAs
            if isinstance(verify, str):
                raise OSError(f"Connection cannot be verified with default trusted CAs. "
                              f"Please provide correct path to a CA_BUNDLE file or directory with "
                              f"certificates of trusted CAs. Error: {e}")

            # forced verify to True
            elif verify:
                raise e

            # default
            elif verify is None:
                verify = 'False'
                kw.update({'verify': False})
                res = func(*args, **kw)

            # disabled verify
            else:
                raise e

        return res

    return wrapper


@set_verify_for_requests
def get(url, params=None, **kwargs):
    r"""Sends a GET request.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return requests.get(url=url, params=params, **kwargs)


@set_verify_for_requests
def options(url, **kwargs):
    r"""Sends an OPTIONS request.

    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return requests.options(url=url, **kwargs)


@set_verify_for_requests
def head(url, **kwargs):
    r"""Sends a HEAD request.

    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes. If
        `allow_redirects` is not provided, it will be set to `False` (as
        opposed to the default :meth:`request` behavior).
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return requests.head(url=url, **kwargs)


@set_verify_for_requests
def post(url, data=None, json=None, **kwargs):
    r"""Sends a POST request.

    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) json data to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return requests.post(url=url, data=data, json=json, **kwargs)


@set_verify_for_requests
def put(url, data=None, **kwargs):
    r"""Sends a PUT request.

    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) json data to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return requests.put(url=url, data=data, **kwargs)


@set_verify_for_requests
def patch(url, data=None, **kwargs):
    r"""Sends a PATCH request.

    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) json data to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return requests.patch(url=url, data=data, **kwargs)


@set_verify_for_requests
def delete(url, **kwargs):
    r"""Sends a DELETE request.

    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return requests.delete(url=url, **kwargs)

class Session(requests.Session):
    """A Requests session.

    Provides cookie persistence, connection-pooling, and configuration.

    Basic Usage::

      >>> import requests
      >>> s = requests.Session()
      >>> s.get('https://httpbin.org/get')
      <Response [200]>

    Or as a context manager::

      >>> with requests.Session() as s:
      ...     s.get('https://httpbin.org/get')
      <Response [200]>
    """
    def __init__(self):
        requests.Session.__init__(self)

    @set_verify_for_requests
    def request(self, method, url, **kwargs):
        """Constructs a :class:`Request <Request>`, prepares it and sends it.
        Returns :class:`Response <Response>` object.

        :param method: method for the new :class:`Request` object.
        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :type timeout: float or tuple
        :param allow_redirects: (optional) Set to True by default.
        :type allow_redirects: bool
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair.
        :rtype: requests.Response
        """
        kwargs['method'] = method
        kwargs['url'] = url
        return super().request(**kwargs)


def session():
    """
    Returns a :class:`Session` for context-management.

    .. deprecated:: 1.0.0

        This method has been deprecated since version 1.0.0 and is only kept for
        backwards compatibility. New code should use :class:`~requests.sessions.Session`
        to create a session. This may be removed at a future date.

    :rtype: Session
    """
    return Session()


"""
EnvHandler for OpenAI gym interface
"""
#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import abc

import gym


class EnvHandler(gym.Env):
    """
    Base class for Environment Handler of Reinforcement learning algorithms
    """

    def __init__(self, data=None, env_config=None):
        """
        Initializes an `EnvHandler` object

        :param config: Start state configuration of environment
        :type config: `dict`
        """

    @abc.abstractmethod
    def step(self, action):
        """
        Run one timestep of the environment's dynamics. When end of
        episode is reached, you are responsible for calling `reset()`
        to reset this environment's state.
        Accepts an action and returns a tuple (observation, reward, done, info).
        Args:
            action (object): an action provided by the agent
        Returns:
            observation (object): agent's observation of the current environment
            reward (float) : amount of reward returned after previous action
            done (bool): whether the episode has ended, in which case further
             step() calls will return undefined results
            info (dict): contains auxiliary diagnostic information
            (helpful for debugging, and sometimes learning)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def reset(self):
        """
        Resets the state of the environment and returns an initial observation.
       Returns:
           observation (object): the initial observation.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def render(self, mode='human'):
        """
        Render one frame of the environment
        """
        raise NotImplementedError



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.tensorflow_artifact_loader import TensorflowArtifactLoader
from ibm_watson_machine_learning.libs.repo.util.library_imports import LibraryChecker
from ibm_watson_machine_learning.libs.repo.base_constants import *

lib_checker = LibraryChecker()

class TensorflowPipelineModelLoader(TensorflowArtifactLoader):
    """
        Returns Tensorflow Runtime  instance associated with this model artifact.

        :return: TensorflowArtifactRunTime instance
        :rtype: TensorflowRuntimeArtifact
        """
    def load_model(self,session=None,tags=None):
        return(self.model_instance(session,tags))



    def model_instance(self,session=None,tags=None):
        lib_checker.check_lib(TENSORFLOW)
        import tensorflow as tf
        if session is None:
            if '2.1.0' in tf.__version__:
                session = tf.compat.v1.Session(graph=tf.compat.v1.Graph())
            else:
                session = tf.Session(graph=tf.Graph())
        elif session is not None:
            if '2.1.0' in tf.__version__ and not isinstance(session, tf.compat.v1.Session):
                raise TypeError("sess should be of type : %s" % tf.compat.v1.Session)
            elif not isinstance(session,tf.Session):
                raise TypeError("sess should be of type : %s" % tf.Session)

        if tags is None:
            if '2.1.0' in tf.__version__:
                tags = [tf.compat.v1.saved_model.tag_constants.SERVING]
            else:
                tags = [tf.saved_model.tag_constants.SERVING]
        elif tags is not None:
            if not isinstance(tags,list):
                raise TypeError("tags should be of type : %s" % list)


        return self.load(session,tags)



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re


class HyperParametersExperiments(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, name=None, double_values=None, int_values=None, string_values=None, double_range=None, int_range=None):
        """
        HyperParametersExperiments - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'name': 'str',
            'double_values': 'list[float]',
            'int_values': 'list[int]',
            'string_values': 'list[str]',
            'double_range': 'HyperParametersExperimentsDoubleRange',
            'int_range': 'HyperParametersExperimentsIntRange'
        }

        self.attribute_map = {
            'name': 'name',
            'double_values': 'double_values',
            'int_values': 'int_values',
            'string_values': 'string_values',
            'double_range': 'double_range',
            'int_range': 'int_range'
        }

        self._name = name
        self._double_values = double_values
        self._int_values = int_values
        self._string_values = string_values
        self._double_range = double_range
        self._int_range = int_range

    @property
    def name(self):
        """
        Gets the name of this HyperParametersExperiments.
        Name of the parameter which we want to tune

        :return: The name of this HyperParametersExperiments.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this HyperParametersExperiments.
        Name of the parameter which we want to tune

        :param name: The name of this HyperParametersExperiments.
        :type: str
        """

        self._name = name

    @property
    def double_values(self):
        """
        Gets the double_values of this HyperParametersExperiments.


        :return: The double_values of this HyperParametersExperiments.
        :rtype: list[float]
        """
        return self._double_values

    @double_values.setter
    def double_values(self, double_values):
        """
        Sets the double_values of this HyperParametersExperiments.


        :param double_values: The double_values of this HyperParametersExperiments.
        :type: list[float]
        """

        self._double_values = double_values

    @property
    def int_values(self):
        """
        Gets the int_values of this HyperParametersExperiments.


        :return: The int_values of this HyperParametersExperiments.
        :rtype: list[int]
        """
        return self._int_values

    @int_values.setter
    def int_values(self, int_values):
        """
        Sets the int_values of this HyperParametersExperiments.


        :param int_values: The int_values of this HyperParametersExperiments.
        :type: list[int]
        """

        self._int_values = int_values

    @property
    def string_values(self):
        """
        Gets the string_values of this HyperParametersExperiments.


        :return: The string_values of this HyperParametersExperiments.
        :rtype: list[str]
        """
        return self._string_values

    @string_values.setter
    def string_values(self, string_values):
        """
        Sets the string_values of this HyperParametersExperiments.


        :param string_values: The string_values of this HyperParametersExperiments.
        :type: list[str]
        """

        self._string_values = string_values

    @property
    def double_range(self):
        """
        Gets the double_range of this HyperParametersExperiments.


        :return: The double_range of this HyperParametersExperiments.
        :rtype: HyperParametersExperimentsDoubleRange
        """
        return self._double_range

    @double_range.setter
    def double_range(self, double_range):
        """
        Sets the double_range of this HyperParametersExperiments.


        :param double_range: The double_range of this HyperParametersExperiments.
        :type: HyperParametersExperimentsDoubleRange
        """

        self._double_range = double_range

    @property
    def int_range(self):
        """
        Gets the int_range of this HyperParametersExperiments.


        :return: The int_range of this HyperParametersExperiments.
        :rtype: HyperParametersExperimentsIntRange
        """
        return self._int_range

    @int_range.setter
    def int_range(self, int_range):
        """
        Sets the int_range of this HyperParametersExperiments.


        :param int_range: The int_range of this HyperParametersExperiments.
        :type: HyperParametersExperimentsIntRange
        """

        self._int_range = int_range

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re


class MetaObjectMetadata(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, guid=None, href=None, created_at=None, modified_at=None, id=None, name=None, owner=None,
                 space_id=None, project_id=None):
        """
        MetaObjectMetadata - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'guid': 'str',
            'href': 'str',
            'created_at': 'datetime',
            'modified_at': 'datetime',
            'id': 'str',
            'name': 'str',
            'owner': 'str',
            'space_id': 'str',
            'project_id': 'str'
        }

        self.attribute_map = {
            'guid': 'guid',
            'href': 'href',
            'created_at': 'created_at',
            'modified_at': 'modified_at',
            'id': 'id',
            'name': 'name',
            'owner': 'owner',
            'space_id': 'space_id',
            'project_id': 'project_id'

        }

        self._guid = guid
        self._href = href
        self._created_at = created_at
        self._modified_at = modified_at

    @property
    def guid(self):
        """
        Gets the guid of this MetaObjectMetadata.


        :return: The guid of this MetaObjectMetadata.
        :rtype: str
        """
        return self._guid

    @guid.setter
    def guid(self, guid):
        """
        Sets the guid of this MetaObjectMetadata.


        :param guid: The guid of this MetaObjectMetadata.
        :type: str
        """

        self._guid = guid

    @property
    def href(self):
        """
        Gets the href of this MetaObjectMetadata.


        :return: The href of this MetaObjectMetadata.
        :rtype: str
        """
        return self._href

    @href.setter
    def href(self, href):
        """
        Sets the href of this MetaObjectMetadata.


        :param href: The href of this MetaObjectMetadata.
        :type: str
        """

        self._href = href

    @property
    def created_at(self):
        """
        Gets the created_at of this MetaObjectMetadata.


        :return: The created_at of this MetaObjectMetadata.
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """
        Sets the created_at of this MetaObjectMetadata.


        :param created_at: The created_at of this MetaObjectMetadata.
        :type: datetime
        """

        self._created_at = created_at

    @property
    def modified_at(self):
        """
        Gets the modified_at of this MetaObjectMetadata.


        :return: The modified_at of this MetaObjectMetadata.
        :rtype: datetime
        """
        return self._modified_at

    @modified_at.setter
    def modified_at(self, modified_at):
        """
        Sets the modified_at of this MetaObjectMetadata.


        :param modified_at: The modified_at of this MetaObjectMetadata.
        :type: datetime
        """

        self._modified_at = modified_at

    @property
    def id(self):
        """
        Gets the id of this MetaObjectMetadata.


        :return: The id of this MetaObjectMetadata.
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """
        Sets the id of this MetaObjectMetadata.


        :param id: The guid of this MetaObjectMetadata.
        :type: str
        """

        self._id = id

    @property
    def name(self):
        """
        Gets the name of this MetaObjectMetadata.


        :return: The name of this MetaObjectMetadata.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this MetaObjectMetadata.


        :param name: The name of this MetaObjectMetadata.
        :type: str
        """

        self._name = name

    @property
    def owner(self):
        """
        Gets the owner of this MetaObjectMetadata.


        :return: The owner of this MetaObjectMetadata.
        :rtype: str
        """
        return self.owner

    @owner.setter
    def owner(self, owner):
        """
        Sets the owner of this MetaObjectMetadata.


        :param owner: The owner of this MetaObjectMetadata.
        :type: str
        """

        self.owner = owner

    @property
    def space_id(self):
        """
        Gets the space_id of this MetaObjectMetadata.


        :return: The space_id of this MetaObjectMetadata.
        :rtype: str
        """
        return self.space_id

    @space_id.setter
    def space_id(self, space_id):
        """
        Sets the space_id of this MetaObjectMetadata.


        :param space_id: The space_id of this MetaObjectMetadata.
        :type: str
        """

        self.space_id = space_id

    @property
    def project_id(self):
        """
        Gets the project_id of this MetaObjectMetadata.


        :return: The project_id of this MetaObjectMetadata.
        :rtype: str
        """
        return self.project_id

    @project_id.setter
    def project_id(self, project_id):
        """
        Sets the project_id of this MetaObjectMetadata.


        :param project_id: The project_id of this MetaObjectMetadata.
        :type: str
        """
        self.project_id = project_id

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

class _Singleton(type):
    """ A metaclass that creates a Singleton base class when called. """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class BaseSingleton(_Singleton('SingletonMeta', (object,), {})):
    pass


#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2019- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.metanames import VolumeMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure
import os
import time
import shlex
import subprocess
_DEFAULT_LIST_LENGTH = 50


class Volume(WMLResource):
    """Store and manage volume assets."""

    ConfigurationMetaNames = VolumeMetaNames()
    """MetaNames for volume assets creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP

    def get_details(self, volume_id):
        """Get volume details.

        :param volume_id: Unique ID of the volume
        :type volume_id: str

        :return: metadata of the volume details
        :rtype: dict

        **Example**

        .. code-block:: python

            volume_details = client.volumes.get_details(volume_id)

        """
        Volume._validate_type(volume_id, u'volume_id', str, True)

        params = {'addon_type': 'volumes',
                  'include_service_status': True
                  }

        response = requests.get(self._client.service_instance._href_definitions.volume_href(volume_id),
                                headers=self._client._get_headers(zen=True))

        if response.status_code == 200:
            return response.json()
        else:
            print(response.status_code, response.text)
            raise WMLClientError("Failed to Get the volume details. Try again.")

    def create(self, meta_props):
        """Create a volume asset.

        :param meta_props: metadata of the volume asset
        :type meta_props: dict

        :return: metadata of the created volume details
        :rtype: dict

        **Examples**

        Provision new PVC volume:

        .. code-block:: python

            metadata = {
                client.volumes.ConfigurationMetaNames.NAME: 'volume-for-wml-test',
                client.volumes.ConfigurationMetaNames.NAMESPACE: 'wmldev2',
                client.volumes.ConfigurationMetaNames.STORAGE_CLASS: 'nfs-client'
                client.volumes.ConfigurationMetaNames.STORAGE_SIZE: "2G"
            }

            asset_details = client.volumes.store(meta_props=metadata)

        Provision an existing PVC volume:

        .. code-block:: python

            metadata = {
                client.volumes.ConfigurationMetaNames.NAME: 'volume-for-wml-test',
                client.volumes.ConfigurationMetaNames.NAMESPACE: 'wmldev2',
                client.volumes.ConfigurationMetaNames.EXISTING_PVC_NAME: 'volume-for-wml-test'
            }

            asset_details = client.volumes.store(meta_props=metadata)

        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Failed to create volume. It is supported only for CP4D 3.5")

        volume_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client
        )

        create_meta = {}
        if self.ConfigurationMetaNames.EXISTING_PVC_NAME in meta_props and \
            meta_props[self.ConfigurationMetaNames.EXISTING_PVC_NAME] is not None:
            if self.ConfigurationMetaNames.STORAGE_CLASS in meta_props and \
                    meta_props[self.ConfigurationMetaNames.STORAGE_CLASS] is not None:
                raise WMLClientError("Failed while creating volume. Either provide EXISTING_PVC_NAME to create a volume using existing volume or"
                                     "provide STORAGE_CLASS and STORAGE_SIZE for new volume creation")
            else:
                create_meta.update({ "existing_pvc_name": meta_props[self.ConfigurationMetaNames.EXISTING_PVC_NAME]})
        else:
            if self.ConfigurationMetaNames.STORAGE_CLASS in meta_props and \
               meta_props[self.ConfigurationMetaNames.STORAGE_CLASS] is not None:
               if self.ConfigurationMetaNames.STORAGE_SIZE in meta_props and \
                       meta_props[self.ConfigurationMetaNames.STORAGE_SIZE] is not None:
                   create_meta.update({"storageClass": meta_props[self.ConfigurationMetaNames.STORAGE_CLASS]})
                   create_meta.update({"storageSize": meta_props[self.ConfigurationMetaNames.STORAGE_SIZE]})
               else:
                   raise WMLClientError("Failed to create volume. Missing input STORAGE_SIZE" )

        if self.ConfigurationMetaNames.EXISTING_PVC_NAME in meta_props and meta_props[self.ConfigurationMetaNames.EXISTING_PVC_NAME] is not None:
            input_meta = {
                "addon_type":"volumes",
                "addon_version":"-",
                "create_arguments":{
                    "metadata":create_meta
                },
                "namespace":meta_props[self.ConfigurationMetaNames.NAMESPACE],
                "display_name":meta_props[self.ConfigurationMetaNames.NAME]
            }
        else:
            input_meta = {
                "addon_type": "volumes",
                "addon_version": "-",
                "create_arguments": {
                    "metadata": create_meta
                },
                "namespace": meta_props[self.ConfigurationMetaNames.NAMESPACE],
                "display_name": meta_props[self.ConfigurationMetaNames.NAME]
            }
        creation_response = {}
        try:
            if self._client.CLOUD_PLATFORM_SPACES:
                creation_response = requests.post(
                    self._client.service_instance._href_definitions.volumes_href(),
                    headers=self._client._get_headers(zen=True),
                    json=input_meta
                )

            else:
                creation_response = requests.post(self._client.service_instance._href_definitions.volumes_href(),
                        headers=self._client._get_headers(zen=True),
                        json=input_meta
                    )
            if creation_response.status_code == 200:
                volume_id_details = creation_response.json()  # messy details returned for backward compability
                import copy
                volume_details = copy.deepcopy(input_meta)
                volume_details.update(volume_id_details)
                actual_details = self.get_details(self.get_id(volume_id_details))
                volume_details.update(actual_details)
                return volume_details
            else:
                print(creation_response.status_code, creation_response.text)
                raise WMLClientError("Failed to create a volume. Try again.")
        except Exception as e:
            print("Exception: ", {e})
            raise WMLClientError("Failed to create a volume. Try again.")

    def start(self, name, wait_for_available=False):
        """Start the volume service.

        :param name: unique name of the volume to be started
        :type name: str

        :param wait_for_available: flag indicating if method should wait until volume service is available
        :type wait_for_available: bool

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.volumes.start(volume_name)

        """

        WMLResource._chk_and_block_create_update_for_python36(self)
        if not self._client.ICP_PLATFORM_SPACES and not self._client.CLOUD_PLATFORM_SPACES:
            raise WMLClientError("Volume APIs are not supported. It is supported only for CP4D 3.5")

        if (self._client.ICP_45 or self._client.ICP_46 or self._client.ICP_47) and '::' not in name:
            raise WMLClientError('Invalid name to start volume. Correct volume name format: `<namespace>::<name>`. Retrieve the correct name using `client.volumes.get_name(client.volumes.get_details(volume_id))` command.')

        start_url = self._client.service_instance._href_definitions.volume_service_href(name)
        # Start the volume  service
        start_data = {}
        try:
            start_data = {}
            creation_response = requests.post(
                start_url,
                headers=self._client._get_headers(zen=True),
                json=start_data
            )
            if creation_response.status_code == 200:
                print("Volume Service started")
                if wait_for_available:
                    retries = 0
                    volume_status = False
                    while True and retries < 60 and not volume_status:
                        volume_status = self.get_volume_status(name)
                        time.sleep(5)
                        retries += 1
                    if not volume_status:
                        print("Volume Service has been started, but it is not available yet."
                              "Check volume availability using get_volume_status method.")

            elif creation_response.status_code == 500:
                print("Failed to start the volume. Make sure volume is in running with status RUNNING or UNKNOW and then re-try")
            else:
                print(creation_response.status_code, creation_response.text)
                raise WMLClientError("Failed to start the file to  volume. Try again.")
        except Exception as e:
            print("Exception:", {e})
            raise WMLClientError("Failed to start the file to  volume. Try again.")

    def get_volume_status(self, name):
        """Monitor a volume's file server status.

        :param name: name of the volume to retrieve status for
        :type name: str

        :return: status of the volume (True if volume is available, otherwise False)
        :rtype: bool

        **Example**

        .. code-block:: python

            client.volumes.get_volume_status(volume_name)

        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        if not self._client.ICP_PLATFORM_SPACES and not self._client.CLOUD_PLATFORM_SPACES:
            raise WMLClientError("Volume APIs are not supported. It is supported only for CP4D 3.5")

        if (self._client.ICP_45 or self._client.ICP_46 or self._client.ICP_47) and '::' not in name:
            raise WMLClientError('Invalid name to start volume. Correct volume name format: `<namespace>::<name>`. Retrieve the correct name using `client.volumes.get_name(client.volumes.get_details(volume_id))` command.')

        monitor_url = self._client.service_instance._href_definitions.volume_monitor_href(name)
        try:
            monitor_response = requests.get(
                monitor_url,
                headers=self._client._get_headers(zen=True)
            )
            if monitor_response.status_code == 200:
                return True
            elif monitor_response.status_code == 502:
                return False
            else:
                print(monitor_response.status_code, monitor_response.text)
                raise WMLClientError("Cannot retrieve status of the volume.")
        except Exception as e:
            print("Exception:", {e})
            raise WMLClientError("Cannot retrieve status of the volume.")

    def upload_file(self, name,  file_path):
        """Upload the data file into stored volume.

        :param name: unique name of the stored volume
        :type name: str
        :param file_path: file to be uploaded into the volume
        :type file_path: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.volumes.upload_file('testA', 'DRUG.csv')

        """
        WMLResource._chk_and_block_create_update_for_python36(self)

        if not self._client.ICP_PLATFORM_SPACES and not self._client.CLOUD_PLATFORM_SPACES:
            raise WMLClientError("Volume APIs are not supported. It is supported only for CP4D 3.5 and above.")

        header_input = self._client._get_headers(zen=True)
        zen_token = header_input.get('Authorization')

        filename_to_upload = file_path.split('/')[-1]
        upload_url_file = self._client.service_instance._href_definitions.volume_upload_href(name) + filename_to_upload
        cmd_str = 'curl -k  -X PUT "' + upload_url_file + '"' + "  -H 'Content-Type: multipart/form-data' -H 'Authorization: " + zen_token + \
                  "' -F upFile='@" + file_path + "'"
        args = shlex.split(cmd_str)
        upload_response = subprocess.run(args, capture_output=True, text=True)
        if upload_response.returncode == 0:
            import json
            try:
                cmd_output = json.loads(upload_response.stdout)
                print(cmd_output.get('message'))
                return "SUCCESS"
            except Exception as e:
                print(upload_response.returncode, upload_response.stdout)
                print("Failed to upload the file to  volume. Try again.")
                return "FAILED"
        else:
            print(upload_response.returncode, upload_response.stdout, upload_response.stderr)
            print("Failed to upload the file to  volume. Try again.")
            return "FAILED"

    def list(self, return_as_df=True):
        """Print stored volumes in a table format. If limit is set to None there will be only first 50 records shown.

        :param return_as_df: determinate if table should be returned as pandas.DataFrame object, default: True
        :type return_as_df: bool, optional

        :return: pandas.DataFrame with listed volumes or None if return_as_df is False
        :rtype: pandas.DataFrame or None

        **Example**

        .. code-block:: python

            client.volumes.list()
        """

        href = self._client.service_instance._href_definitions.volumes_href()
        params = {}
        params.update({'addon_type': 'volumes'})

        response = requests.get(href, params=params, headers=self._client._get_headers(zen=True))

        asset_details = self._handle_response(200, u'list volumes', response)
        asset_list = asset_details.get('service_instances')
        volume_values = [
            (m[u'display_name'],
             m[u'id'],
             m['provision_status']) for m in asset_list]

        table = self._list(volume_values, [u'NAME', u'ID', u'PROVISION_STATUS'], None, _DEFAULT_LIST_LENGTH)

        if return_as_df:
            return  table


    @staticmethod
    def get_id(volume_details):
        """Get unique Id of stored volume details.

        :param volume_details: metadata of the stored volume details
        :type volume_details: dict

        :return: unique Id of stored volume asset
        :rtype: str

        **Example**

        .. code-block:: python

            volume_uid = client.volumes.get_id(volume_details)

        """

        Volume._validate_type(volume_details, u'asset_details', object, True)
        if 'service_instance' in volume_details and  volume_details.get('service_instance') is not None:
            vol_details = volume_details.get('service_instance')
            return WMLResource._get_required_element_from_dict(vol_details, u'volume_assets_details',
                                                               [u'id'])
        else:
            return WMLResource._get_required_element_from_dict(volume_details, u'volume_assets_details',
                                                           [u'id'])

    @staticmethod
    def get_name(volume_details):
        """Get unique name of stored volume asset.

        :params volume_details: metadata of the stored volume asset
        :type volume_details: dict

        :return: unique name of stored volume asset
        :rtype: str

        **Example**

        .. code-block:: python

            volume_name = client.volumes.get_name(asset_details)

        """
        Volume._validate_type(volume_details, u'asset_details', object, True)
        if 'service_instance' in volume_details and  volume_details.get('service_instance') is not None:
            vol_details = volume_details.get('service_instance')
            return WMLResource._get_required_element_from_dict(vol_details, u'volume_assets_details',
                                                               [u'display_name'])
        else:
            return WMLResource._get_required_element_from_dict(volume_details, u'volume_assets_details',
                                                           [u'display_name'])

    def delete(self, volume_id):
        """Delete a volume.

        :param volume_id: unique ID of the volume
        :type volume_id: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.volumes.delete(volume_id)

        """
        Volume._validate_type(volume_id, u'asset_uid', str, True)
        if (not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES):
            raise WMLClientError(u'Volume API is not supported.')

        response = requests.delete(self._client.service_instance._href_definitions.volume_href(volume_id),
                                   headers=self._client._get_headers(zen=True))

        if response.status_code == 200 or response.status_code == 204:
            print("Successfully deleted volume service.")
            return "SUCCESS"
        else:
            print("Failed to delete volume.")
            print(response.status_code, response.text)
            return "FAILED"

    def stop(self, volume_name):
        """Stop the volume service.

        :param volume_name: unique name of the volume
        :type volume_name: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.volumes.stop(volume_name)

        """
        Volume._validate_type(volume_name, u'asset_uid', str, True)
        if (not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES):
            raise WMLClientError(u'Volume API is not supported.')

        response = requests.delete(self._client.service_instance._href_definitions.volume_service_href(volume_name),
                                   headers=self._client._get_headers(zen=True))

        if response.status_code == 200:
            print("Successfully stopped volume service.")
            return "SUCCESS"
        else:
            print(response.status_code, response.text)
            return "FAILED"




#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from .spark_artifact_loader import SparkArtifactLoader


class SparkPipelineLoader(SparkArtifactLoader):
    """
    Returns pipeline instance associated with this pipeline artifact.

    :return: pipeline
    :rtype: pyspark.ml.Pipeline
    """
    def pipeline_instance(self):
        return self.load()



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
from cryptography import x509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa

from ibmfl.crypto.crypto_exceptions import *
from ibmfl.crypto.infra.crypto_cert_int import CryptoCert


class CryptoCertRsa(CryptoCert):
    """
    This class implements the interface for certificate verification functions using x509 and RSA.
    """

    def __init__(self, ca_cert_file_path: str, my_cert_file_path: str, **kwargs):
        super(CryptoCertRsa, self).__init__(ca_cert_file_path, my_cert_file_path)
        if ca_cert_file_path is None:
            raise KeyDistributionInputException("ca_cert_file_path is None")
        else:
            with open(ca_cert_file_path, "rb") as ca_cert_file:
                self.ca_cert = x509.load_pem_x509_certificate(ca_cert_file.read())
            self.ca_public_key = self.ca_cert.public_key()
            if not isinstance(self.ca_public_key, rsa.RSAPublicKey):
                raise KeyDistributionInputException(
                    "Invalid type of public key in certificate=" + repr(type(self.ca_public_key))
                )
            with open(my_cert_file_path, "rb") as my_cert_file:
                self.my_cert = x509.load_pem_x509_certificate(my_cert_file.read())
            vermy, _ = self.verify_cert_signature(self.my_cert)
            if not vermy:
                raise KeyDistributionVerificationException("Invalid certificate self.my_cert=" + repr(self.my_cert))

    def verify_cert_signature(self, certificate):
        if isinstance(certificate, bytes):
            check_cert = x509.load_pem_x509_certificate(certificate)
        elif isinstance(certificate, str):
            with open(certificate, "rb") as check_cert_file:
                check_cert = x509.load_pem_x509_certificate(check_cert_file.read())
        elif not isinstance(certificate, x509.Certificate):
            raise KeyDistributionInputException("Invalid type of certificate=" + repr(type(certificate)))
        else:
            check_cert = certificate
        try:
            self.ca_public_key.verify(
                check_cert.signature,
                check_cert.tbs_certificate_bytes,
                padding.PKCS1v15(),
                check_cert.signature_hash_algorithm,
            )
        except InvalidSignature:
            return (False, None)
        return (True, check_cert.public_key())

    def get_my_cert(self, ret_type: str):
        if ret_type == "pem":
            ret = self.my_cert.public_bytes(serialization.Encoding.PEM)
        elif ret_type == "obj":
            ret = self.my_cert
        else:
            raise KeyDistributionInputException("Invalid value of ret_type=" + repr(ret_type))
        return ret



"""
Module acting as a bridge between Server and ProtocolHandler
Routes for all the requests received by the server are redirected
to designated handler in PH using routers
"""
#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import logging
from parse import parse
from ibmfl.exceptions import DuplicateRouteException


logger = logging.getLogger(__name__)


class Router(object):
    """
    Container used to add and match a group of routes.
    """

    def __init__(self):
        self.routes = {}

    def add_route(self, path, handler):
        """
        Add path, handler combination in a dictionary.
        Throws DuplicateRouteException when route is repeated

        :param path: special pre defined /assigned identifier which will be \
            used by the client to reach specific handler in the PH
        :param handler: method which needs to handle the request \
            carrying the special identifier
        :raises `DuplicateRouteException`: exception is raised \
            when path is added multiple times
        """
        if path not in self.routes:
            self.routes[path] = handler
        else:
            raise DuplicateRouteException

    def add_routes(self, routes):
        """
        Add multiple path, handler combinations in routes dictionary.
        Throws DuplicateRouteException when route is repeated

        :param routes: list of path, handler combination
        :raises `DuplicateRouteException`: exception is raised when \
        path is added multiple times
        """
        for route, handler in routes.items():
            self.add_route(route, handler)

    def get_handler(self, request_path):
        """
        Gets handler for path specified in the request

        :param request_path: path or identifier provided in \
            the request message
        :return: a handler which was assigned when Router object \
            was created
        """
        for path, handler in self.routes.items():
            parse_result = parse(path, request_path)
            if parse_result is not None:
                return handler, parse_result.named

        return None, None



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2018- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import re

class GenericArchiveFrameworkCheck(object):
    @staticmethod
    def is_archive_framework(name):
        if (re.match('spss-modeler', name) is not None) or (re.match('pmml', name) is not None) \
          or (re.match('caffe', name) is not None) or (re.match('caffe2', name) is not None) \
          or (re.match('torch',name) is not None) \
          or (re.match('pytorch', name) is not None) or (re.match('blueconnect', name) is not None) \
          or (re.match('mxnet', name) is not None) or (re.match('theano', name) is not None) \
          or (re.match('darknet', name) is not None):
          return True
        else:
         return False



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import copy
from typing import List, Union
from warnings import warn

from ibm_watson_machine_learning.preprocessing import DataJoinGraph
from ibm_watson_machine_learning.utils.autoai.enums import (
    TShirtSize, ClassificationAlgorithms, RegressionAlgorithms, ForecastingAlgorithms, PredictionType, Metrics, \
    Transformers, DataConnectionTypes, PipelineTypes, PositiveLabelClass, ClassificationAlgorithmsCP4D,
    RegressionAlgorithmsCP4D, ForecastingAlgorithmsCP4D, SamplingTypes, ImputationStrategy, ForecastingPipelineTypes,
    TimeseriesAnomalyPredictionPipelineTypes, TimeseriesAnomalyPredictionAlgorithms)
from ibm_watson_machine_learning.utils.autoai.errors import LocalInstanceButRemoteParameter, MissingPositiveLabel, \
    NonForecastPredictionColumnMissing, ForecastPredictionColumnsMissing, ForecastingCannotBeRunAsLocalScenario, \
    TSNotSupported, TSADNotSupported, ParamOutOfRange, ImputationListNotSupported, \
    MissingEstimatorForExistingBatchedEstimator, TimeseriesAnomalyPredictionFeatureColumnsMissing, \
    TimeseriesAnomalyPredictionCannotBeRunAsLocalScenario, TimeseriesAnomalyPredictionUnsupportedMetric
from ibm_watson_machine_learning.utils.autoai.utils import check_dependencies_versions, \
    validate_additional_params_for_optimizer, validate_optimizer_enum_values, \
    translate_imputation_string_strategy_to_enum, translate_estimator_string_to_enum, \
    translate_batched_estimator_string_to_enum
from ibm_watson_machine_learning.workspace import WorkSpace
from ibm_watson_machine_learning.wml_client_error import ForbiddenActionForGitBasedProject, WMLClientError
from ibm_watson_machine_learning.messages.messages import Messages
from .engines import WMLEngine
from .optimizers import LocalAutoPipelines, RemoteAutoPipelines
from .runs import AutoPipelinesRuns, LocalAutoPipelinesRuns
from ..base_experiment.base_experiment import BaseExperiment

__all__ = [
    "AutoAI"
]


class AutoAI(BaseExperiment):
    """AutoAI class for pipeline models optimization automation.

    :param wml_credentials: credentials to Watson Machine Learning instance
    :type wml_credentials: dict

    :param project_id: ID of the Watson Studio project
    :type project_id: str, optional

    :param space_id: ID of the Watson Studio Space
    :type space_id: str, optional

    :param verify: user can pass as verify one of following:

        - the path to a CA_BUNDLE file
        - the path of directory with certificates of trusted CAs
        - `True` - default path to truststore will be taken
        - `False` - no verification will be made
    :type verify: bool or str, optional

    **Example**

    .. code-block:: python

        from ibm_watson_machine_learning.experiment import AutoAI

        experiment = AutoAI(
            wml_credentials={
                "apikey": "...",
                "iam_apikey_description": "...",
                "iam_apikey_name": "...",
                "iam_role_crn": "...",
                "iam_serviceid_crn": "...",
                "instance_id": "...",
                "url": "https://us-south.ml.cloud.ibm.com"
            },
            project_id="...",
            space_id="...")
    """
    # note: initialization of AutoAI enums as class properties

    # note: Enums with estimators can be overwritten  in _init based on environment type (CPD or Cloud)
    ClassificationAlgorithms = ClassificationAlgorithms
    RegressionAlgorithms = RegressionAlgorithms
    ForecastingAlgorithms = ForecastingAlgorithms
    # end note
    TShirtSize = TShirtSize
    PredictionType = PredictionType
    Metrics = Metrics
    Transformers = Transformers
    DataConnectionTypes = DataConnectionTypes
    PipelineTypes = PipelineTypes
    SamplingTypes = SamplingTypes

    def __init__(self,
                 wml_credentials: Union[dict, 'WorkSpace'] = None,
                 project_id: str = None,
                 space_id: str = None,
                 verify=None) -> None:
        # note: as workspace is not clear enough to understand, there is a possibility to use pure
        # wml credentials with project and space IDs, but in addition we
        # leave a possibility to use a previous WorkSpace implementation, it could be passed as a first argument
        if wml_credentials is None:
            self._workspace = None
            self.runs = LocalAutoPipelinesRuns()

        else:
            if isinstance(wml_credentials, WorkSpace):
                self._workspace = wml_credentials
            else:
                self._workspace = WorkSpace(wml_credentials=wml_credentials.copy(),
                                            project_id=project_id,
                                            space_id=space_id,
                                            verify=verify)

            self.project_id = self._workspace.project_id
            self.space_id = self._workspace.space_id
            self.runs = AutoPipelinesRuns(engine=WMLEngine(self._workspace))
            self.runs._workspace = self._workspace

        #self._block_autoai_on_git_based_project()

        self._init_estimator_enums()

        self._20_class_limit_removal_test = False
        # --- end note

    def runs(self, *, filter: str) -> Union['AutoPipelinesRuns', 'LocalAutoPipelinesRuns']:
        """Get the historical runs but with WML Pipeline name filter (for remote scenario).
        Get the historical runs but with experiment name filter (for local scenario).

        :param filter: WML Pipeline name to filter the historical runs or experiment name to filter
            the local historical runs
        :type filter: str

        :return: object managing the list of runs
        :rtype: AutoPipelinesRuns or LocalAutoPipelinesRuns

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI

            experiment = AutoAI(...)
            experiment.runs(filter='Test').list()
        """

        if self._workspace is None:
            return LocalAutoPipelinesRuns(filter=filter)

        else:
            return AutoPipelinesRuns(engine=WMLEngine(self._workspace.wml_client), filter=filter)

    def optimizer(self,
                  name: str,
                  *,
                  prediction_type: 'PredictionType',
                  prediction_column: str = None,
                  prediction_columns: List[str] = None,
                  timestamp_column_name: str = None,
                  scoring: 'Metrics' = None,
                  desc: str = None,
                  test_size: float = None,  # deprecated
                  holdout_size: float = None,
                  max_number_of_estimators: int = None,
                  train_sample_rows_test_size: float = None,
                  include_only_estimators: List[Union['ClassificationAlgorithms', 'RegressionAlgorithms',
                                                      'ForecastingAlgorithms', 'TimeseriesAnomalyPredictionAlgorithms']] = None,
                  daub_include_only_estimators: List[Union['ClassificationAlgorithms', 'RegressionAlgorithms']] = None,  # deprecated
                  include_batched_ensemble_estimators: List[Union['BatchedClassificationAlgorithms',
                                                                  'BatchedRegressionAlgorithms']] = None,
                  backtest_num: int = None,
                  lookback_window: int = None,
                  forecast_window: int = None,
                  backtest_gap_length: int = None,
                  feature_columns: List[str] = None,
                  pipeline_types: List[Union['ForecastingPipelineTypes', 'TimeseriesAnomalyPredictionPipelineTypes']] = None,
                  supporting_features_at_forecast: bool = None,
                  cognito_transform_names: List['Transformers'] = None,
                  data_join_graph: 'DataJoinGraph' = None,
                  csv_separator: Union[List[str], str] = ',',
                  excel_sheet: Union[str, int] = None,
                  encoding: str = 'utf-8',
                  positive_label: str = None,
                  data_join_only: bool = False,
                  drop_duplicates: bool = True,
                  outliers_columns: list = None,
                  text_processing: bool = None,
                  word2vec_feature_number: int = None,
                  daub_give_priority_to_runtime: float = None,
                  fairness_info: dict = None,
                  sampling_type: 'SamplingTypes' = None,
                  sample_size_limit: int = None,
                  sample_rows_limit: int = None,
                  sample_percentage_limit: float = None,
                  n_parallel_data_connections: int = None,
                  number_of_batch_rows: int = None,
                  categorical_imputation_strategy: ImputationStrategy = None,
                  numerical_imputation_strategy: ImputationStrategy = None,
                  numerical_imputation_value: float = None,
                  imputation_threshold: float = None,
                  retrain_on_holdout: bool = None,
                  categorical_columns: list = None,
                  numerical_columns: list = None,
                  test_data_csv_separator: Union[List[str], str] = ',',
                  test_data_excel_sheet: str = None,
                  test_data_encoding: str = 'utf-8',
                  confidence_level: float = None,
                  incremental_learning: bool = None,
                  early_stop_enabled: bool = None,
                  early_stop_window_size: int = None,
                  **kwargs) -> Union['RemoteAutoPipelines', 'LocalAutoPipelines']:
        """
        Initialize an AutoAi optimizer.

        :param name: name for the AutoPipelines
        :type name: str

        :param prediction_type: type of the prediction
        :type prediction_type: PredictionType

        :param prediction_column: name of the target/label column, required for `multiclass`, `binary` and `regression`
            prediction types
        :type prediction_column: str, optional

        :param prediction_columns: names of the target/label columns, required for `forecasting` prediction type
        :type prediction_columns: list[str], optional

        :param timestamp_column_name: name of timestamp column for time series forecasting
        :type timestamp_column_name: str, optional

        :param scoring: type of the metric to optimize with, not used for forecasting
        :type scoring: Metrics, optional

        :param desc: description
        :type desc: str, optional

        :param test_size: deprecated, use `holdout_size` instead

        :param holdout_size: percentage of the entire dataset to leave as a holdout
        :type holdout_size: float, optional

        :param max_number_of_estimators: maximum number (top-K ranked by DAUB model selection)
            of the selected algorithm, or estimator types, for example `LGBMClassifierEstimator`,
            `XGBoostClassifierEstimator`, or `LogisticRegressionEstimator` to use in pipeline composition,
            the default is `None` that means the true default value will be determined by
            the internal different algorithms, where only the highest ranked by model selection algorithm type is used
        :type max_number_of_estimators: int, optional

        :param train_sample_rows_test_size: training data sampling percentage
        :type train_sample_rows_test_size: float, optional

        :param daub_include_only_estimators: deprecated, use `include_only_estimators` instead
        
        :param include_batched_ensemble_estimators: list of batched ensemble estimators to include 
            in computation process, see: AutoAI.BatchedClassificationAlgorithms, AutoAI.BatchedRegressionAlgorithms
        :type include_batched_ensemble_estimators: 
            list[BatchedClassificationAlgorithms or BatchedRegressionAlgorithms], optional

        :param include_only_estimators: list of estimators to include in computation process, see:
            AutoAI.ClassificationAlgorithms, AutoAI.RegressionAlgorithms or AutoAI.ForecastingAlgorithms
        :type include_only_estimators: List[ClassificationAlgorithms or RegressionAlgorithms or ForecastingAlgorithms]], optional

        :param backtest_num: number of backtests used for forecasting prediction type, default value: 4,
            value from range [0, 20]
        :type backtest_num: int, optional

        :param lookback_window: length of lookback window used for forecasting prediction type,
            default value: 10, if set to -1 lookback window will be auto-detected
        :type lookback_window: int, optional

        :param forecast_window: length of forecast window used for forecasting prediction type, default value: 1,
            value from range [1, 60]
        :type forecast_window: int, optional

        :param backtest_gap_length: gap between backtests used for forecasting prediction type,
            default value: 0, value from range [0, data length / 4]
        :type backtest_gap_length: int, optional

        :param feature_columns: list of feature columns used for forecasting prediction type,
            may contain target column and/or supporting feature columns, list of columns to be detected whether there are anomalies for timeseries anomaly prediction type
        :type feature_columns: list[str], optional

        :param pipeline_types: list of pipeline types to be used for forecasting or timeseries anomaly prediction type
        :type pipeline_types: list[ForecastingPipelineTypes or TimeseriesAnomalyPredictionPipelineTypes], optional

        :param supporting_features_at_forecast: enables usage of future supporting feature values during forecast
        :type supporting_features_at_forecast: bool, optional

        :param cognito_transform_names: list of transformers to include in the feature enginnering computation process,
            see: AutoAI.Transformers
        :type cognito_transform_names: list[Transformers], optional

        :param csv_separator: the separator, or list of separators to try for separating columns in a CSV file,
            not used if the file_name is not a CSV file, default is ','
        :type csv_separator: list[str] or str, optional

        :param excel_sheet: name of the excel sheet to use, only applicable when xlsx file is an input,
            support for number of the sheet is deprecated, by default first sheet is used
        :type excel_sheet: str, optional

        :param encoding: encoding type for CSV training file
        :type encoding: str, optional

        :param positive_label: the positive class to report when binary classification, when multiclass or regression,
            this will be ignored
        :type positive_label: str, optional

        :param t_shirt_size: the size of the remote AutoAI POD instance (computing resources),
            only applicable to a remote scenario, see: AutoAI.TShirtSize
        :type t_shirt_size: TShirtSize, optional

        :param data_join_graph: a graph object with definition of join structure for multiple input data sources,
            data preprocess step for multiple files
        :type data_join_graph: DataJoinGraph, optional

        :param data_join_only: if `True` only preprocessing will be executed
        :type data_join_only: bool, optional

        :param drop_duplicates: if `True` duplicated rows in data will be removed before further processing
        :type drop_duplicates: bool, optional

        :param outliers_columns: replace outliers with NaN using IQR method for specified columns. By default,
            turned ON for regression learning_type and target column. To turn OFF empty list of columns must be passed
        :type outliers_columns: list, optional

        :param text_processing: if `True` text processing will be enabled, applicable only on Cloud
        :type text_processing: bool, optional

        :param word2vec_feature_number: number of features which will be generated from text column,
            will be applied only if `text_processing` is `True`, if `None` the default value will be taken
        :type word2vec_feature_number: int, optional

        :param daub_give_priority_to_runtime: the importance of run time over score for pipelines ranking,
            can take values between 0 and 5, if set to 0.0 only score is used,
            if set to 1 equally score and runtime are used, if set to value higher than 1
            the runtime gets higher importance over score
        :type daub_give_priority_to_runtime: float, optional

        :param fairness_info: dictionary that specifies metadata needed for measuring fairness,
            it contains three key values: `favorable_labels`, `unfavorable_labels` and `protected_attributes`,
            the `favorable_labels` attribute indicates that when the class column contains one of the value from list,
            that is considered a positive outcome, the `unfavorable_labels` is oposite to the `favorable_labels`
            and is obligatory for regression learning type, a protected attribute is a list of features that partition
            the population into groups whose outcome should have parity, if protected attribute is empty list
            then automatic detection of protected attributes will be run,
            if `fairness_info` is passed then fairness metric will be calculated
        :type fairness_info: fairness_info

        :param n_parallel_data_connections: number of maximum parallel connection to data source,
            supported only for IBM Cloud Pak® for Data 4.0.1 and above
        :type n_parallel_data_connections: int, optional

        :param categorical_imputation_strategy: missing values imputation strategy for categorical columns

            Possible values (only non-forecasting scenario):

            - ImputationStrategy.MEAN
            - ImputationStrategy.MEDIAN
            - ImputationStrategy.MOST_FREQUENT (default)

        :type categorical_imputation_strategy: ImputationStrategy, optional


        :param numerical_imputation_strategy: missing values imputation strategy for numerical columns

            Possible values (non-forecasting scenario):

            - ImputationStrategy.MEAN
            - ImputationStrategy.MEDIAN (default)
            - ImputationStrategy.MOST_FREQUENT

            Possible values (forecasting scenario):

            - ImputationStrategy.MEAN
            - ImputationStrategy.MEDIAN
            - ImputationStrategy.BEST_OF_DEFAULT_IMPUTERS (default)
            - ImputationStrategy.VALUE
            - ImputationStrategy.FLATTEN_ITERATIVE
            - ImputationStrategy.LINEAR
            - ImputationStrategy.CUBIC
            - ImputationStrategy.PREVIOUS
            - ImputationStrategy.NEXT
            - ImputationStrategy.NO_IMPUTATION

        :param numerical_imputation_value: value for filling missing values if numerical_imputation_strategy
            is set to ImputationStrategy.VALUE, for forecasting only
        :type numerical_imputation_value: float, optional

        :param imputation_threshold: maximum threshold of missing values imputation, for forecasting only
        :type imputation_threshold: float, optional

        :param retrain_on_holdout: if True final pipelines will be train also on holdout data
        :type retrain_on_holdout: bool, optional

        :param categorical_columns: list of columns names that must be treated as categorical
        :type categorical_columns: list, optional

        :param numerical_columns: list of columns names that must be treated as numerical
        :type numerical_columns: list, optional

        :param sampling_type: type of sampling data for training, one of SamplingTypes enum values,
            default is SamplingTypes.FIRST_N_RECORDS, supported only for IBM Cloud Pak® for Data 4.0.1 and above
        :type sampling_type: str, optional

        :param sample_size_limit: the size of sample upper bound (in bytes). The default value is 1GB,
            supported only for IBM Cloud Pak® for Data 4.5 and above
        :type sample_size_limit: int, optional

        :param sample_rows_limit: the size of sample upper bound (in rows),
            supported only for IBM Cloud Pak® for Data 4.6 and above
        :type sample_rows_limit: int, optional

        :param sample_percentage_limit: the size of sample upper bound (as fraction of dataset size),
            supported only for IBM Cloud Pak® for Data 4.6 and above
        :type sample_percentage_limit: float, optional

        :param number_of_batch_rows: number of rows to read in each batch when reading from flight connection
        :type number_of_batch_rows: int, optional

        :param test_data_csv_separator: the separator, or list of separators to try for separating
            columns in a CSV user-defined holdout/test file, not used if the file_name is not a CSV file,
            default is ','
        :type test_data_csv_separator: list[str] or str, optional

        :param test_data_excel_sheet: name of the excel sheet to use for user-defined holdout/test data, 
            only use when xlsx file is an test, dataset file, by default first sheet is used
        :type test_data_excel_sheet: str or int, optional

        :param test_data_encoding: encoding type for CSV user-defined holdout/test file
        :type test_data_encoding: str, optional

        :param confidence_level: when the pipeline "PointwiseBoundedHoltWinters" or "PointwiseBoundedBATS" is used, 
            the prediction interval is calculated at a given confidence_level to decide if a data record 
            is an anomaly or not, optional for timeseries anomaly prediction
        :type confidence_level: float, optional

        :param incremental_learning: triggers incremental learning process for supported pipelines
        :type incremental_learning: bool, optional

        :param early_stop_enabled: enables early stop for incremental learning process
        :type early_stop_enabled: bool, optional

        :param early_stop_window_size: the number of iterations without score improvements before training stop
        :type early_stop_window_size: int, optional

        :return: RemoteAutoPipelines or LocalAutoPipelines, depends on how you initialize the AutoAI object
        :rtype: RemoteAutoPipelines or LocalAutoPipelines

        **Examples**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI
            experiment = AutoAI(...)

            fairness_info = {
                       "protected_attributes": [
                           {"feature": "Sex", "reference_group": ['male'], "monitored_group": ['female']},
                           {"feature": "Age", "reference_group": [[50,60]], "monitored_group": [[18, 49]]}
                       ],
                       "favorable_labels": ["No Risk"],
                       "unfavorable_labels": ["Risk"],
                       }

            optimizer = experiment.optimizer(
                   name="name of the optimizer.",
                   prediction_type=AutoAI.PredictionType.BINARY,
                   prediction_column="y",
                   scoring=AutoAI.Metrics.ROC_AUC_SCORE,
                   desc="Some description.",
                   holdout_size=0.1,
                   max_number_of_estimators=1,
                   fairness_info= fairness_info,
                   cognito_transform_names=[AutoAI.Transformers.SUM,AutoAI.Transformers.MAX],
                   train_sample_rows_test_size=1,
                   include_only_estimators=[AutoAI.ClassificationAlgorithms.LGBM, AutoAI.ClassificationAlgorithms.XGB],
                   t_shirt_size=AutoAI.TShirtSize.L
               )

            optimizer = experiment.optimizer(
                   name="name of the optimizer.",
                   prediction_type=AutoAI.PredictionType.MULTICLASS,
                   prediction_column="y",
                   scoring=AutoAI.Metrics.ROC_AUC_SCORE,
                   desc="Some description.",
               )
        """
        # note: convert `timeseries` type to PredictionType.FORECASTING:
        if prediction_type == 'timeseries':
            prediction_type = PredictionType.FORECASTING


        if prediction_type != PredictionType.FORECASTING and retrain_on_holdout is None:
            retrain_on_holdout = True

        # Deprecation of excel_sheet as number:
        if isinstance(excel_sheet, int):
            warn(
                message="Support for excel sheet as number of the sheet (int) is deprecated! Please set excel sheet with name of the sheet.")

        if data_join_graph is not None:
            if self._workspace.wml_client.ICP_46 or self._workspace.wml_client.ICP_47:
                raise WMLClientError(Messages.get_message(message_id="obm_removal_message_cpd"))
            elif self._workspace.wml_client.ICP:
                print(Messages.get_message(message_id="obm_deprecation_message_cpd"))
            else:
                raise WMLClientError(Messages.get_message(message_id="obm_removal_message_cloud"))

        if prediction_type == PredictionType.FORECASTING and self._workspace.wml_client.ICP and \
                (self._workspace.wml_client.wml_credentials['version'].startswith('2.5') or \
                        self._workspace.wml_client.wml_credentials['version'].startswith('3.0') or \
                        self._workspace.wml_client.wml_credentials['version'].startswith('3.5')):
            raise TSNotSupported()

        if prediction_type == PredictionType.TIMESERIES_ANOMALY_PREDICTION and self._workspace.wml_client.ICP and \
                self._workspace.wml_client.wml_credentials['version'].startswith(('2.5', '3.0', '3.5', '4.0', '4.5', '4.6')):
            raise TSADNotSupported()

        if prediction_type in (PredictionType.FORECASTING, 'timeseries'):
            if not numerical_imputation_strategy and type(numerical_imputation_strategy) is not list:
                numerical_imputation_strategy = ImputationStrategy.BEST_OF_DEFAULT_IMPUTERS
            elif not numerical_imputation_strategy and type(numerical_imputation_strategy) is list:
                numerical_imputation_strategy = ImputationStrategy.NO_IMPUTATION

            if prediction_column is not None or prediction_columns is None:
                raise ForecastPredictionColumnsMissing()
        elif prediction_type == PredictionType.TIMESERIES_ANOMALY_PREDICTION:
            if feature_columns is None or prediction_column is not None or prediction_columns is not None:
                raise TimeseriesAnomalyPredictionFeatureColumnsMissing()
            if scoring is not None and scoring not in (
                    Metrics.F1_SCORE, Metrics.ROC_AUC_SCORE, Metrics.AVERAGE_PRECISION_SCORE, Metrics.PRECISION_SCORE,
                    Metrics.RECALL_SCORE):
                raise TimeseriesAnomalyPredictionUnsupportedMetric(scoring)
        else:
            if prediction_column is None or prediction_columns is not None:
                raise NonForecastPredictionColumnMissing(prediction_type)

        if test_size:
            print('Note: Using `test_size` is deprecated. Use `holdout_size` instead.')
            if not holdout_size:
                holdout_size = test_size
            test_size = None

        if daub_include_only_estimators:
            print('Note: Using `daub_include_only_estimators` is deprecated. Use `include_only_estimators` instead.')
            if not include_only_estimators:
                include_only_estimators = daub_include_only_estimators
            daub_include_only_estimators = None

        if train_sample_rows_test_size and (self._workspace.wml_client.ICP_46 or self._workspace.wml_client.ICP_47):
            print('Note: Using `train_sample_rows_test_size` is deprecated.'
                  'Use either `sample_rows_limit` or `sample_percentage_limit` instead.')
            if not sample_rows_limit and not sample_percentage_limit:
                if type(train_sample_rows_test_size) is float and train_sample_rows_test_size <= 1:
                    print('Value of `train_sample_rows_test_size` parameter'
                          'will be passed as `sample_percentage_limit`.')
                    sample_percentage_limit = train_sample_rows_test_size
                elif int(train_sample_rows_test_size) == train_sample_rows_test_size and train_sample_rows_test_size > 1:
                    print('Value of `train_sample_rows_test_size` parameter'
                          'will be passed as `sample_rows_limit`.')
                    sample_rows_limit = int(train_sample_rows_test_size)
                train_sample_rows_test_size = None
            elif sample_rows_limit or sample_percentage_limit:
                print('Parameter `train_sample_rows_test_size` will be ignored.')
                train_sample_rows_test_size = None

        def translate_str_imputation_param(x):
            if type(x) is list and prediction_type != PredictionType.FORECASTING:
                raise ImputationListNotSupported()

            if type(x) == str or (type(x) == list and type(x[0]) == str):
                return translate_imputation_string_strategy_to_enum(x, prediction_type)
            else:
                return x

        def translate_str_include_only_estimators_param(x):
            return [translate_estimator_string_to_enum(estimator) for estimator in x]

        def translate_str_include_batched_ensemble_estimators_param(x):
            return [translate_batched_estimator_string_to_enum(estimator) for estimator in x]

        def translate_str_pipeline_types_param(x):
            if prediction_type == PredictionType.TIMESERIES_ANOMALY_PREDICTION:
                return [TimeseriesAnomalyPredictionPipelineTypes(pipeline_type) for pipeline_type in x]
            else:
                return [ForecastingPipelineTypes(pipeline_type) for pipeline_type in x]

        categorical_imputation_strategy = translate_str_imputation_param(categorical_imputation_strategy)
        numerical_imputation_strategy = translate_str_imputation_param(numerical_imputation_strategy)
        include_only_estimators = translate_str_include_only_estimators_param(include_only_estimators) if include_only_estimators else None
        include_batched_ensemble_estimators = translate_str_include_batched_ensemble_estimators_param(include_batched_ensemble_estimators) if include_batched_ensemble_estimators else None
        pipeline_types = translate_str_pipeline_types_param(pipeline_types) if pipeline_types != None else None

        if include_batched_ensemble_estimators:
            for batched_estimator in include_batched_ensemble_estimators:
                basic_estimator_str = batched_estimator.value.split("(")[1][:-1]
                basic_estimator = translate_estimator_string_to_enum(basic_estimator_str)
                if include_only_estimators is not None and basic_estimator not in include_only_estimators:
                    raise MissingEstimatorForExistingBatchedEstimator(batched_estimator, basic_estimator)

        validate_optimizer_enum_values(
            prediction_type=prediction_type,
            daub_include_only_estimators=daub_include_only_estimators,
            include_only_estimators=include_only_estimators,
            include_batched_ensemble_estimators=include_batched_ensemble_estimators,
            cognito_transform_names=cognito_transform_names,
            imputation_strategies=[x for y in list(filter(None, [categorical_imputation_strategy, numerical_imputation_strategy])) for x in (y if type(y) is list else [y])],
            scoring=scoring,
            t_shirt_size=kwargs.get("t_shirt_size", TShirtSize.M),
            is_cpd=self._workspace.wml_client.ICP
        )

        if daub_give_priority_to_runtime is not None:
            if daub_give_priority_to_runtime < 0.0 or daub_give_priority_to_runtime > 5.0:
                raise ParamOutOfRange('daub_give_priority_to_runtime', daub_give_priority_to_runtime, 0.0, 5.0)

        if data_join_graph:
            data_join_graph.problem_type = prediction_type
            data_join_graph.target_column = prediction_column

        if (prediction_type == PredictionType.BINARY and scoring in vars(PositiveLabelClass).values()
                and positive_label is None):
            raise MissingPositiveLabel(scoring, reason=f"\"{scoring}\" needs a \"positive_label\" "
                                                       f"parameter to be defined when used with binary classification.")

        if self._workspace is None and kwargs.get('t_shirt_size'):
            raise LocalInstanceButRemoteParameter(
                "t_shirt_size",
                reason="During LocalOptimizer initialization, \"t_shirt_size\" parameter was provided. "
                       "\"t_shirt_size\" parameter is only applicable to the RemoteOptimizer instance."
            )
        elif self._workspace is None:
            if prediction_type == PredictionType.FORECASTING:
                raise ForecastingCannotBeRunAsLocalScenario()
            if prediction_type == PredictionType.TIMESERIES_ANOMALY_PREDICTION:
                raise TimeseriesAnomalyPredictionCannotBeRunAsLocalScenario()

            reduced_kwargs = copy.copy(kwargs)

            for n in ['_force_local_scenario']:
                if n in reduced_kwargs:
                    del reduced_kwargs[n]

            validate_additional_params_for_optimizer(reduced_kwargs)

            return LocalAutoPipelines(
                name=name,
                prediction_type='classification' if prediction_type in ['binary', 'multiclass'] else prediction_type,
                prediction_column=prediction_column,
                scoring=scoring,
                desc=desc,
                holdout_size=holdout_size,
                max_num_daub_ensembles=max_number_of_estimators,
                train_sample_rows_test_size=train_sample_rows_test_size,
                include_only_estimators=include_only_estimators,
                include_batched_ensemble_estimators=include_batched_ensemble_estimators,
                cognito_transform_names=cognito_transform_names,
                positive_label=positive_label,
                _force_local_scenario=kwargs.get('_force_local_scenario', False),
                **reduced_kwargs
            )

        else:
            reduced_kwargs = copy.copy(kwargs)

            if max_number_of_estimators is None:
                max_number_of_estimators = reduced_kwargs.get('max_num_daub_ensembles', None)

            for n in ['t_shirt_size', 'notebooks', 'autoai_pod_version', 'obm_pod_version', 'max_num_daub_ensembles']:
                if n in reduced_kwargs:
                    del reduced_kwargs[n]

            validate_additional_params_for_optimizer(reduced_kwargs)

            engine = WMLEngine(self._workspace)

            if self._20_class_limit_removal_test:
                engine._20_class_limit_removal_test = True

            optimizer = RemoteAutoPipelines(
                name=name,
                prediction_type=prediction_type,
                prediction_column=prediction_column,
                prediction_columns=prediction_columns,
                timestamp_column_name=timestamp_column_name,
                scoring=scoring,
                desc=desc,
                holdout_size=holdout_size,
                max_num_daub_ensembles=max_number_of_estimators,
                t_shirt_size=self._workspace.restrict_pod_size(t_shirt_size=kwargs.get(
                    't_shirt_size', TShirtSize.M if self._workspace.wml_client.ICP else TShirtSize.L)
                ),
                train_sample_rows_test_size=train_sample_rows_test_size,
                include_only_estimators=include_only_estimators,
                include_batched_ensemble_estimators=include_batched_ensemble_estimators,
                backtest_num=backtest_num,
                lookback_window=lookback_window,
                forecast_window=forecast_window,
                backtest_gap_length=backtest_gap_length,
                cognito_transform_names=cognito_transform_names,
                data_join_graph=data_join_graph,
                drop_duplicates=drop_duplicates,
                outliers_columns=outliers_columns,
                text_processing=text_processing,
                word2vec_feature_number=word2vec_feature_number,
                csv_separator=csv_separator,
                excel_sheet=excel_sheet,
                encoding=encoding,
                positive_label=positive_label,
                data_join_only=data_join_only,
                engine=engine,
                daub_give_priority_to_runtime=daub_give_priority_to_runtime,
                notebooks=kwargs.get('notebooks', True),
                autoai_pod_version=kwargs.get('autoai_pod_version', None),
                obm_pod_version=kwargs.get('obm_pod_version', None),
                fairness_info=fairness_info,
                sampling_type=sampling_type,
                sample_size_limit=sample_size_limit,
                sample_rows_limit=sample_rows_limit,
                sample_percentage_limit=sample_percentage_limit,
                number_of_batch_rows=number_of_batch_rows,
                n_parallel_data_connections=n_parallel_data_connections,
                categorical_imputation_strategy=categorical_imputation_strategy,
                numerical_imputation_strategy=numerical_imputation_strategy,
                numerical_imputation_value=numerical_imputation_value,
                imputation_threshold=imputation_threshold,
                retrain_on_holdout=retrain_on_holdout,
                numerical_columns=numerical_columns,
                categorical_columns=categorical_columns,
                feature_columns=feature_columns,
                pipeline_types=pipeline_types,
                supporting_features_at_forecast=supporting_features_at_forecast,
                test_data_csv_separator=test_data_csv_separator,
                test_data_excel_sheet=test_data_excel_sheet,
                test_data_encoding=test_data_encoding,
                confidence_level=confidence_level,
                incremental_learning=incremental_learning,
                early_stop_enabled=early_stop_enabled,
                early_stop_window_size=early_stop_window_size,
                **reduced_kwargs
            )
            optimizer._workspace = self._workspace
            return optimizer

    def _init_estimator_enums(self):
        if self._workspace and self._workspace.wml_client.ICP:
            self.ClassificationAlgorithms = ClassificationAlgorithmsCP4D
            self.RegressionAlgorithms = RegressionAlgorithmsCP4D
            self.ForecastingAlgorithms = ForecastingAlgorithmsCP4D
        else:
            self.ClassificationAlgorithms = ClassificationAlgorithms
            self.RegressionAlgorithms = RegressionAlgorithms
            self.ForecastingAlgorithms = ForecastingAlgorithms

    def _block_autoai_on_git_based_project(self):
        """Raises ForbiddenActionForGitBasedProject error for AutoAI experiments on git based project.
        It can be disabled by setting environment variable ENABLE_AUTOAI to 'true'
        """
        from os import environ

        if self._workspace:
            if getattr(self._workspace.wml_client, 'project_type', None) == 'local_git_storage' \
                    and environ.get('ENABLE_AUTOAI', 'false').lower() == 'false':
                raise ForbiddenActionForGitBasedProject(reason="Creating AutoAI experiment is not supported for git based project.")





#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from ibm_watson_machine_learning.libs.repo.swagger_client.apis.repository_api import RepositoryApi
from ibm_watson_machine_learning.libs.repo.swagger_client.rest import ApiException

import ibm_watson_machine_learning._wrappers.requests as requests
import certifi, os

try:
    import urllib3
except ImportError:
    raise ImportError('urllib3 is missing')

try:
    # for python3
    from urllib.parse import urlencode
except ImportError:
    # for python2
    from urllib import urlencode

http = urllib3.PoolManager(
    cert_reqs='CERT_REQUIRED',
    ca_certs=certifi.where()
)


class MLRepositoryApi(RepositoryApi):
    """
    MLRepositoryApi extends RepositoryApi.
    """
    def __init__(self, api_client):
        super(MLRepositoryApi, self).__init__(api_client)

    def upload_pipeline_version(self, pipeline_id, version_id, content_stream):
        os.environ["DEPLOYMENT_PLATFORM"] = "private"
        if 'DEPLOYMENT_PLATFORM' in os.environ and os.environ['DEPLOYMENT_PLATFORM'] == 'private':
            r = requests.put('{}/v3/ml_assets/training_definitions/{}/versions/{}/content'.format(
                self.api_client.repository_path,
                pipeline_id,
                version_id
            ),
                headers=self.api_client.default_headers,
                data=content_stream
            )
            if r.status_code != 200:
                raise ApiException(r.status_code, r.text)
        else:
            r = requests.put('{}/v3/ml_assets/training_definitions/{}/versions/{}/content'.format(
                self.api_client.repository_path,
                pipeline_id,
                version_id
            ),
                headers=self.api_client.default_headers,
                data=content_stream
            )
            if r.status_code != 200:
                raise ApiException(r.status_code, r.text)

    def upload_pipeline_model_version(self, model_id, version_id, content_stream, query_param=None):
        if query_param is not None:
            query_parameters = query_param
        else:
            query_parameters = {}


        if 'DEPLOYMENT_PLATFORM' in os.environ and os.environ['DEPLOYMENT_PLATFORM'] == 'private':
            r = requests.put('{}/ml/v4/models/{}/content'.format(
                self.api_client.repository_path,
                model_id

            ),
                headers=self.api_client.default_headers,
                data=content_stream,
                params=query_parameters
            )
            if (r.status_code != 200 and r.status_code !=202):
                raise ApiException(r.status_code, r.text)
        else:
            r = requests.put('{}/ml/v4/models/{}/content'.format(
                self.api_client.repository_path,
                model_id

            ),
                headers=self.api_client.default_headers,
                data=content_stream,
                params=query_parameters
            )

            if (r.status_code != 200 and r.status_code !=202):
                r = requests.delete('{}/ml/v4/models/{}'.format(
                    self.api_client.repository_path,
                    model_id
                ),
                    headers=self.api_client.default_headers,
                    params=query_parameters
                )
                raise ApiException(r.status_code, r.text)

    def upload_pipeline_model_v4_cloud(self, model_id, content_stream, query_param=None, no_delete=None):

        if query_param is not None:
            query_parameters = query_param
        else:
            query_parameters = {}

        r = requests.put('{}/ml/v4/models/{}/content'.format(
            self.api_client.repository_path,
            model_id

        ),
            headers=self.api_client.default_headers,
            data=content_stream,
            params=query_parameters
        )
        if (r.status_code != 201 and r.status_code !=202):
            if no_delete is False or no_delete is None:
                r = requests.delete('{}/ml/v4/models/{}'.format(
                    self.api_client.repository_path,
                    model_id
                ),
                    headers=self.api_client.default_headers,
                    params=query_parameters
                )
            raise ApiException(r.status_code, r.text)

    def get_model_size(self, content_href):
        return(self.v3_ml_assets_models_artifact_size_get(content_href))

    def get_presigned_url(self, content_href):
        return(self.v3_ml_assets_models_artifact_content_get(content_href))

    def get_async_status(self, url):
        return(self.v3_ml_assets_models_async_status_get(url))

    def download_artifact_content_v4_cloud(self, artifact_content_href, presigned_url, accept='application/octet-stream'):
        query_params = {}
        header_params = {"Accept": accept}
        header_1 = {'Content-Type': 'application/gzip'}
        tmp_headers = self.api_client.default_headers.copy()
        tmp_headers.update(header_params)
        tmp_headers.update(header_1)
        if presigned_url == 'true':
            splitOne = artifact_content_href.rsplit('?')
            splitTwo = splitOne[0].rsplit('/', 1)
            baseUrl = splitTwo[0]
            url = splitTwo[1]
            splitThree = splitOne[1].rsplit('&')
            query_params_list = {}
            for elem in splitThree:
                key_value = elem.split('=')
                query_params_list[key_value[0]]=key_value[1].replace('%2F', '/')


            artifact_content_href = (baseUrl+"/"+url)
            query_params = query_params_list
        try:
            return self.api_client.download_file_v4(
                   artifact_content_href,
                   presigned_url,
                   query_params,
                   tmp_headers
               )

        except ApiException as ex:
            raise ApiException(500, 'Internal server error during download content for: {}. Error message: {}'.format(artifact_content_href, ex))

    def download_artifact_content(self, artifact_content_href, presigned_url, accept='application/json'):
        query_params = {}
        header_params = {'Accept': accept}
        if presigned_url == 'true':
            splitOne = artifact_content_href.rsplit('?')
            splitTwo = splitOne[0].rsplit('/', 1)
            baseUrl = splitTwo[0]
            url = splitTwo[1]
            splitThree = splitOne[1].rsplit('&')
            query_params_list = {}
            for elem in splitThree:
                key_value = elem.split('=')
                query_params_list[key_value[0]]=key_value[1].replace('%2F', '/')


            artifact_content_href = (baseUrl+"/"+url)
            query_params = query_params_list
        try:
            return self.api_client.download_file(
                   artifact_content_href,
                   presigned_url,
                   query_params,
                   header_params
               )

        except ApiException as ex:
            raise ApiException(500, 'Internal server error during download content for: {}. Error message: {}'.format(artifact_content_href, ex))
        #else:
        #    raise ValueError('The artifact content href: {} is not within the client host: {}'
        #                     .format(artifact_content_href,
        #                             self.api_client.repository_path))



    def upload_function_revision(self, function_id, revision_id, content_stream):

        os.environ["DEPLOYMENT_PLATFORM"] = "private"
        if 'DEPLOYMENT_PLATFORM' in os.environ and os.environ['DEPLOYMENT_PLATFORM'] == 'private':
            r = requests.put('{}/v4/functions/{}/revisions/{}/content'.format(
                self.api_client.repository_path,
                function_id,
                revision_id
            ),
                headers=self.api_client.default_headers,
                data=content_stream
            )
            if r.status_code != 200:
                raise ApiException(r.status_code, r.text)
        else:
            r = requests.put('{}/v4/functions/{}/revisions/{}/content'.format(
                self.api_client.repository_path,
                function_id,
                revision_id
            ),
                headers=self.api_client.default_headers,
                data=content_stream
            )
            if r.status_code != 200:
                raise ApiException(r.status_code, r.text)

    def upload_libraries(self, library_id, content_stream):

        if 'DEPLOYMENT_PLATFORM' in os.environ and os.environ['DEPLOYMENT_PLATFORM'] == 'private':
            r = requests.put('{}/v4/libraries/{}/content'.format(
                self.api_client.repository_path,
                library_id),
                headers=self.api_client.default_headers,
                data=content_stream
            )
            if r.status_code != 200 and r.status_code !=202:
                raise ApiException(r.status_code, r.text)
        else:
            r = requests.put('{}/v4/libraries/{}/content'.format(
                self.api_client.repository_path,
                library_id),
                headers=self.api_client.default_headers,
                data=content_stream
            )
            if r.status_code != 200 and r.status_code !=202:
                raise ApiException(r.status_code, r.text)

    def upload_runtimes(self, runtimespec_id, content_stream):
        header_params = {'Content-Type': 'text/plain'}
        tmp_headers = self.api_client.default_headers.copy()
        tmp_headers.update(header_params)

        if 'DEPLOYMENT_PLATFORM' in os.environ and os.environ['DEPLOYMENT_PLATFORM'] == 'private':
            r = requests.put('{}/v4/runtimes/{}/content'.format(
                self.api_client.repository_path,
                runtimespec_id
            ),
                headers=tmp_headers,
                data=content_stream
            )
            if r.status_code != 200 and r.status_code !=202:
                raise ApiException(r.status_code, r.text)
        else:
            r = requests.put('{}/v4/runtimes/{}/content'.format(
                self.api_client.repository_path,
                runtimespec_id
            ),
                headers=tmp_headers,
                data=content_stream
            )
            if r.status_code != 200 and r.status_code !=202:
                raise ApiException(r.status_code, r.text)
